#include <math.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <stdlib.h>
#include <assert.h>
#include "additional.h"

#define PITI 0.0000001
#define TOUPITIPITI 1.0e-40

static double double_abs(double x)
{
       if(x<0) x = -x;
       return(x);
}

/*****
 * classe matrice
 * classe gerant les matrices 
 *****/

/*****
 * matrice::matrice(int d)
 * constructeur de la classe: renvoie une matrice avec des 0 partout
 *****/
matrice::matrice(int d){
  int i,j;
  dimension = d;
  tab = new (double*) [d];
  for (i=0;i<d;i++){
    tab[i] = new (double) [d];
    for (j=0;j<d;j++){
      tab[i][j]=0;
    }
  }
}

/*****
 * matrice::~matrice()
 * destructeur de la classe
 *****/
matrice::~matrice(){
  int i;
  for (i=0;i<dimension;i++){
    delete [] tab[i];
  }
  delete [] tab;
}


/*****
 * void matrice::affiche(void);
 * Affiche la matrice
 *****/
void matrice::affiche(void){
  int i,j;

  printf("\n");
  for (i=0;i<dimension;i++){
    for (j=0;j<dimension;j++){
      printf("%g ",tab[i][j]);
    }
    printf("\n");
  }
  printf("\n");
}

/*****
 * int matrice::check_inverse(matrice *B)
 * verifie que le receveur et B sont bien inverses
 *****/
int matrice::check_inverse(matrice *B){
  matrice* C=new matrice(dimension);
  matrice* D=new matrice(dimension);
  int res;
  double frob;

  C->putProduct(this, B);
  D->putIdentity();
  
  frob = C->frobeniusDelta(D);

  //printf("frob = %g\n",frob);
  
  res = (frob < PITI) ? 1 : 0;

  delete D;
  delete C;

  return res;
}

/*****
 * void matrice::putIdentity();
 * met I dans la matrice
 *****/
void matrice::putIdentity(){
  int i,j;
  for (i=0;i<dimension;i++){
    for (j=0;j<dimension;j++){
      tab[i][j] = (i==j) ? 1 : 0;
    }
  }
}

/*****
 * int matrice::noZeroInDiagonal(void)
 * renvoie 1 ssi il n'y a aucun zero sur la diagonale
 *****/
int matrice::noZeroInDiagonal(void){
  int i;
  int val=1;
  for (i=0;i<dimension;i++){
    if (tab[i][i]==0) val = 0;
  }
  return val;
}


/*****
 * int matrice::approxZeroInDiagonal(void)
 * renvoie 1 ssi il n'y a aucun chiffre proche du zero
 * sur la diagonale
 *****/
int matrice::approxZeroInDiagonal(void){
  int i;
  int val=1;
  for (i=0;i<dimension;i++){
    if (tab[i][i]<=TOUPITIPITI) val = 0;
  }
  return val;
}


/*****
 * double matrice::frobeniusDelta(matrice *B)
 * renvoie la norme de Frobenius de receveur - B
 *****/
double matrice::frobeniusDelta(matrice *B){
  assert(dimension == B->getDim());

  double res=0;
  int i,j;

  for (i=0;i<dimension;i++)
    for (j=0;j<dimension;j++){
      res+=(tab[i][j]-B->getCoord(i,j))*(tab[i][j]-B->getCoord(i,j));
    }
  res = sqrt(res);

  return res;
}

/*****
 * void matrice::putProduct(matrice* A, matrice* B)
 * met dans le receveur le produit AB
 *****/
void matrice::putProduct(matrice* A, matrice* B){
  assert(dimension == A->getDim());
  assert(dimension == B->getDim());
 
  int i,j,k;
  for (i=0;i<dimension;i++){
    for (j=0;j<dimension;j++){
      tab[i][j]=0;
      for (k=0;k<dimension;k++){
	tab[i][j]+=(A->getCoord(i,k))*(B->getCoord(k,j));
      }
    }
  }
}

#define TOUPITI 1.0e-20

/*****
 * int matrice::ludcmp(int *indx,double *d)
 *****/
int matrice::ludcmp(int *indx,double *d)
{
  int n=dimension;
  int i,imax,j,k;
  double big,dum,sum,temp;
  double *vv;
  
  vv=new (double) [n];
  
  *d=1.0;
  for (i=0;i<n;i++) 
    {
      big=0.0;
      for (j=0;j<n;j++)
	if ((temp=double_abs(tab[i][j])) > big)
	  big=temp;
      if (big == 0.0)
	return 0; /* Singular matrix  */
      vv[i]=1.0/big;
    }
  for (j=0;j<n;j++) 
    {
      for (i=0;i<j;i++) 
	{
	  sum=tab[i][j];
	  for (k=0;k<i;k++)
	    sum -= tab[i][k]*tab[k][j];
	  tab[i][j]=sum;
        }
      big=0.0;
      for (i=j;i<n;i++) 
	{
	  sum=tab[i][j];
	  for (k=0;k<j;k++)
	    sum -= tab[i][k]*tab[k][j];
	  tab[i][j]=sum;
	  if ( (dum=vv[i]*double_abs(sum)) >= big) 
	    {
	      big=dum;
	      imax=i;
            }
        }
      if (j != imax) 
	{
	  for (k=0;k<n;k++) 
	    {
	      dum=tab[imax][k];
	      tab[imax][k]=tab[j][k];
	      tab[j][k]=dum;
            }
	  *d = -(*d);
	  vv[imax]=vv[j];
        }
      indx[j]=imax;
      
      if (tab[j][j] == 0.0)
	tab[j][j]=TOUPITI;
      if (j != n) 
	{
	  dum=1.0/(tab[j][j]);
	  for (i=j+1;i<n;i++)
	    tab[i][j] *= dum;
        }
      

      /* Change made 3/27/98 for robustness */
      /*if ( (tab[j][j]>=0)&&(tab[j][j]<TOUPITI) ) tab[j][j]= TOUPITI;
	if ( (tab[j][j]<0)&&(tab[j][j]>-TOUPITI) ) tab[j][j]= -TOUPITI;
      
	if (j != n-1) 
	{
	dum=1.0/(tab[j][j]);
	for (i=j+1;i<n;i++)
	tab[i][j] *= dum;
        }
      */
    }
  
  delete [] vv;
  
  return(1);
}

#undef TOUPITI

/*****
 * void matrice::lubksb(int *indx,double *b)
 *****/
void matrice::lubksb(int *indx,double *b){
  int n=dimension;
  int i,ii,ip,j;
  double sum;
  
  ii = -1;
  for (i=0;i<n;i++)
    {
      ip=indx[i];
      sum=b[ip];
      b[ip]=b[i];
      if (ii >= 0)
	for (j=ii;j<i;j++)
	  sum -= tab[i][j]*b[j];
      else if (sum)
	ii=i;
      b[i]=sum;
    }
  for (i=n-1;i>=0;i--)
    {
      sum=b[i];
      for (j=i+1;j<n;j++)
	sum -= tab[i][j]*b[j];
      b[i]=sum/tab[i][i];
    }
}

/*****
 * int matrice::inverse(double *det_man, int *det_exp)
 * Inverse le receveur, quelquesoit la matrice
 * Attention: this = matrice entree / sortie
 * *det_man = mantisse du determinant
 * *det_exp = exposant du determinant
 * -> le determinant vaut det_man*10^{det_exp}
 * renvoie 0 si A est singuliere 
 *****/
int matrice::inverse(double *det_man, int *det_exp){
  int  i,j,n=dimension;
  double  d_man;
  int d_exp;
  
  matrice *y = new matrice(n);

  int* indx = new (int) [n];
  double* col = new (double) [n];


  d_exp = 0;
  if(ludcmp(indx,&d_man)) {
    for(j=0; j<n; j++) {
      d_man *= tab[j][j];
      while( double_abs(d_man)>10 ) {
	d_man = d_man/10;
	d_exp++;
      }
      while( (double_abs(d_man)<0.1)&&(double_abs(d_man)>0) ) {
	d_man = d_man*10;
	d_exp--;
      }
    }
    *det_man = d_man;
    *det_exp = d_exp;
    for(j=0; j<n; j++) {
      for(i=0; i<n; i++) col[i]=0.0;
      col[j]=1.0;
      lubksb(indx,col);
      for(i=0; i<n; i++) y->putCoord(i,j,col[i]);
    } 
    
    for(i=0; i<n; i++)
      for(j=0; j<n; j++) tab[i][j]=y->getCoord(i,j);
    return(1);
  }
  else {
    *det_man = 0.0;
    *det_exp = 0;
    return(0);
  }
		     
  delete [] col;
  delete [] indx;
  delete [] y;
}

/*****
 * int matrice::inverse_diagonal(double *det_man, int *det_exp)
 * Inverse le receveur, qui doit etre diagonal (dur !)
 * Attention: this = matrice entree / sortie
 * *det_man = mantisse du determinant
 * *det_exp = exposant du determinant
 * -> le determinant vaut det_man*10^{det_exp}
 * renvoie 0 si A est singuliere 
 *****/
int matrice::inverse_diagonal(double *det_man, int *det_exp){
  int n = dimension;

  int  i;
  double  d_man;
  int d_exp;
  double test_det=1;
  
  d_exp = 0;
  d_man = 1;
  for(i=0; i<n; i++) {
    d_man *= tab[i][i];
    test_det*= tab[i][i];
    while( double_abs(d_man)>10 ) {
      d_man = d_man/10;
      d_exp++;
    }
    while( (double_abs(d_man)<0.1)&&(double_abs(d_man)>0) ) {
      d_man = d_man*10;
      d_exp--;
    }
  }
  *det_man = d_man;
  *det_exp = d_exp;
  
  for(i=0; i<n; i++) {
    if(tab[i][i]!=0) tab[i][i] = 1.0/tab[i][i];
    else {
      *det_man = 0;
      return(1);
    }
  }
 
  //printf("Determinant trouve=%g et %g\n",test_det,exp(log(*det_man)+*det_exp*log(10)));
}


void test_inv(){
  int dim=10;

  matrice* testm = new matrice(dim);
  matrice* ori = new matrice(dim);
  matrice* scratchm = new matrice(dim);
  matrice* I = new matrice(dim);
 
  double *col;
  int* indx;
  double det_man;
  int det_exp, res;

  indx = new (int) [dim];
  col = new (double) [dim];

  int i,j,k;

  for (i=0;i<dim;i++){
    for (j=0;j<dim;j++){
      scratchm->putCoord(i,j,0);
      if (i==j){
	testm->putCoord(i,j,rand()%1000);
	I->putCoord(i,j,1);
      }
      else{
	testm->putCoord(i,j,0);
	I->putCoord(i,j,0);
      }
      ori->putCoord(i,j,testm->getCoord(i,j));
    }
  }

  testm->affiche();
  res = testm->inverse_diagonal(&det_man, &det_exp);
  //res = clust_invert(testm, dim, &det_man, &det_exp, indx, scratchm, col);
  testm->affiche();
  printf("Singular=%d, Man=%g | Exp=%d | Det=%g\n\n",1-res,det_man,det_exp,exp(log(det_man)+det_exp*log(10)));

  scratchm->putProduct(ori,testm);
  scratchm->affiche();

  printf("FrobDelta=%g\n",scratchm->frobeniusDelta(I));
}
