/*****
* cluster-main.cxx
* prg principal
*****/

#include"cluster.h"
//#include"additional.h"
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <stdlib.h>

#ifdef __GNUG__
#include "getopt.h"
#endif

#ifndef GETOPT_H
extern int optind, opterr; // pour getopt()
extern char     *optarg;
#endif

int ddim, ppts, kkth, ggen, bbal, kkex, sste, ssau, ttim, wwei, ccon, iini, ddiag, zzzz;
char ffil[255];
char option_char;

int lire_options(int argc, char * argv[]){

  //valeurs par defaut
  ddim=2;
  ppts=1000;
  kkth=10;
  ggen=0;
  bbal=0;
  kkex=5;
  ssau=1;
  sste=10;
  ttim=512;
  wwei=0;
  ccon=0;
  iini=1;
  ddiag=1;
  zzzz=0;
  strcpy(ffil,"default");

  while ((option_char = getopt (argc, argv, "h:d:n:k:g:b:D:K:I:S:N:T:W:C:Z:")) != EOF) {
    switch (option_char) {
    case 'd' :
      ddim=atoi(optarg);
      break;
    case 'n' :
      ppts=atoi(optarg);
      break;
    case 'k' :
      kkth=atoi(optarg);
      break;
    case 'g' :
      ggen=atoi(optarg);
      break;
    case 'b' :
      bbal=atoi(optarg);
      break;
    case 'D' :
      ddiag=atoi(optarg);
      break;
    case 'K' :
      kkex=atoi(optarg);
      break;
    case 'I' :
      iini=atoi(optarg);
      break;
    case 'S' :
      sste=atoi(optarg);
      break;
    case 'N' :
      ssau=atoi(optarg);
      break;
    case 'T' :
      ttim=atoi(optarg);
      break;
    case 'W' :
      wwei=atoi(optarg);
      break;
    case 'C' :
      ccon=atoi(optarg);
      break;
    case 'Z' :
      zzzz=atoi(optarg);
      break;
    case 'h' :
    default  : cout << "Using Clusters :\n";
      cout << argv[0] << " [option]* string_name\n";
      cout << " Options:\n";
      cout << "  h : display this help\n";
      cout << "  d : dimension of the problem (d>=1, default is 2) \n";
      cout << "  n : number of points (default is 1000) \n";
      cout << "  k : number of theoretical clusters (default is 10) \n";
      cout << "  g : type of clusters (0=Gaussian, \n";
      cout << "                        1=Ring Gaussian, \n";
      cout << "                        2=Grid Gaussian, BIRCH-like, with dimension==2, \n";
      cout << "                        3=0 with overlap reduced, \n";
      cout << "                        4=Hypercubic uniform, default is 0)\n";
      cout << "  b : balance between sampling the theoretical clusters (0=Uniform: all proportions are identical, \n";
      cout << "                                                         1=Skewed random: each proportion is picked at random, \n";
      cout << "                                                         2=Unbalanced: proportions are approximately exponentially decreasing, default is 0)\n";
      cout << "  K : number of experimental clusters (default is 5) \n";
      cout << "  I : type of initialisation (0=Random, 1=Forgy, default is 1) \n";
      cout << "  S : number of running steps (default is 10) \n";
      cout << "  N : #steps between saving each image (default is 1, \n";
      cout << "                                        i.e. each config image is saved) \n";
      cout << "  T : width(=length) of .tif images saving the configs (default is 512) \n";
      cout << "  W : show the weights on the images (0/1, default is 0=no) \n";
      cout << "  C : show the 2D-convex hulls on the images (0/1, default is 0=no) \n";
      cout << "      (please pick 1 only with K Means)\n";
      cout << "  Z : type of algorithm (0=Kmeans, 1=Harmonic Means, 2=EM, 3=Fuzzy C Means, default is Kmeans) \n";
      cout << "  D : diagonal covariance matrices for Gaussian EM (0=No, 1=Yes, default is 1) \n";
      cout << "  --- string_name is a generic name used to save (i) the config images\n";
      cout << "      (by appending [string_name]_image_[#step].tif), (ii) the KMN-Losses\n";
      cout << "      (the filename is [string_name]_params.txt), and (iii) the theoretical\n";
      cout << "      configuration image (the filename is [string_name]_theorique.tif)\n";
      cout << "      the default value for string_name=\"default\"\n";
      cout << "  -Example (cut and paste):\n";
      cout << "            ./cluster.exe -d 2 -n 2000 -k 2 -g 1 -K 16 -b 0 -I 0 -S 50 -N 1 -T 200 -W 1 -C 0 -Z 0 my_test\n";
      return(0);
    }
  }
  if (  optind < argc  ) strcpy(ffil,argv[optind]);

  return 1;
}

int main(int argc, char * argv[]) {
  if(!lire_options(argc,argv)) exit(1);
  printf("\nWelcome !\n\nAdaptive Weighted Clustering (Nock & Nielsen, 2005)\nPlease read the companion README.TXT file\n\nHere are the parameters I read from the command line:\n");
  printf(" -> dimension: %d\n",ddim);
  printf(" -> number of points: %d\n",ppts);
  printf(" -> number of theoretical clusters: %d\n",kkth);
  if (ggen == 0)
    printf(" ----> the clusters are spherical Gaussians (with identity covariance matrix)\n");
  if (ggen == 1)
    printf(" ----> the clusters are ring Gaussians\n");
  if (ggen == 2)
    printf(" ----> the clusters are grid, BIRCH-like Gaussians\n");
  if (ggen == 3)
    printf(" ----> the clusters are spherical Gaussians (with identity covariance matrix), with smaller overlap\n");
  if (ggen == 4)
    printf(" ----> the clusters are hypercubic uniform\n");
  if (bbal == 0)
    printf(" ----> the cluster proportions are identical (each has approx. the same number of points)\n");
  if (bbal == 1)
    printf(" ----> the cluster proportions are random\n");
  if (bbal == 2)
    printf(" ----> the cluster proportions are *Exponentially decreasing* with the cluster number\n");
  printf(" -> number of empirical clusters: %d\n",kkex);
  if (iini == 0)
    printf(" ----> initialisation of empirical centers: average random subsets of data\n");
  if (iini == 1)
    printf(" ----> initialisation of empirical centers: Forgy\n");
  printf(" -> number of clustering steps: %d\n",sste);
  printf(" -> number of steps between saving configuration images (%d x %d pixels each): %d\n",ttim,ttim,ssau);
  if (wwei == 0)
    printf(" ----> weights are not shown on the images\n");
  else
    printf(" ----> weights are shown with circles on the images (the larger a circle radius, the larger the weight)\n");
  if (ccon == 0)
    printf(" ----> convex hulls of clusters are not shown on the images\n");
  else
    printf(" ----> convex hulls of clusters are shown on the images\n");
  if (zzzz == 0)
    printf(" -> Algorithms ran: K Means\n");
  else if (zzzz == 1)
    printf(" -> Algorithms ran: Harmonic Means\n");
  else if (zzzz == 2)
    printf(" -> Algorithms ran: Gaussian Expectation Maximization\n");
  else if (zzzz == 3)
    printf(" -> Algorithms ran: Fuzzy C Means\n");
  if ( (zzzz==2) && (ddiag == 0) )
    printf(" ----> Covariance matrices are arbitrary\n");
  else if ( (zzzz==2) && (ddiag == 1) )
    printf(" ----> Covariance matrices are diagonal (faster and more robust)\n");

  printf("\n\n");

  //test_inv();

  tabPoints* t; 
  t=new tabPoints(ddim,ggen,bbal,ppts,kkth);
  
  if ( (zzzz == 1) || (zzzz == 2) || (zzzz == 3) )
    t->generate(VRAI); 
  else
    t->generate(FAUX);
  //Rescaling data (more robust)
  
  if (ccon == 1){
    if (zzzz > 0){
      printf("Please use convex hulls ONLY with K Means\n");
    }
    assert(zzzz>0);
  }

  clusters* c;
  c=new clusters(t,kkex,ffil,ttim,ssau,wwei,ccon,iini,VRAI);
  
  clusters* cw;
  cw=new clusters(t,kkex,ffil,ttim,ssau,wwei,ccon,iini,FAUX);
  cw->reinitialise_clustering(c); //cw est une copie reinitialisee sur c
 
  c->put_dist(zzzz); 
  cw->put_dist(zzzz);
  //ATTENTION: DIST = 0 <=> KM ; DIST = 1 <=> HM ; DIST = 2 <=> EM
  //DIST = 3 <=> FCM

  if (zzzz == 0){
    cw->Generic_K_Means(sste,0);
    c->Generic_K_Means(sste,2);
  }else if (zzzz == 1){
    cw->Generic_Harmonic_Means(sste,0);
    c->Generic_Harmonic_Means(sste,2);
  }else if (zzzz == 2){
    if (ddiag == 1){
      cw->Generic_Expectation_Maximization(sste,0,VRAI);
      c->Generic_Expectation_Maximization(sste,2,VRAI);
    }else{
      cw->Generic_Expectation_Maximization(sste,0,FAUX);
      c->Generic_Expectation_Maximization(sste,2,FAUX);
    }
  }else if (zzzz == 3){
    cw->Generic_Fuzzy_C_Means(sste,0);
    c->Generic_Fuzzy_C_Means(sste,2);
  }

  c->sauvegarde_costs_against(cw);
  c->sauvegarde_props_against(cw);
  c->sauvegarde_average_props_against(cw);
  c->sauvegarde_missed_against(cw);
  c->sauvegarde_max_loss_against(cw);
}

