/*****
* cluster.cxx : programme gerant la creation d'images et
* les algo de clustering adaptif
* ATTENTION: on suit le schema de Hammerly et Elkan, CIKM'03
*****/

/*****
* declaration des variables importantes du programme (a faire passer en ligne de comm)
*****/

#define max(x,y) ((x)>(y) ? (x) : (y))

////////////////////////////////////////////////////////////////////////////
//You CAN modify the following variables: they are used for graphical
//conventions. However, keep them in their predefined range (indicated)
////////////////////////////////////////////////////////////////////////////

int EM_PUT_ISOVALEURS = 2;  
// in {0,1,2}
// 0 = graphical like Harmonic mean clustering
// (to each cluster is associated a RGB color; the soft membership
//  composes the colors)
// 1 = puts ellipses with isovalues in colors
// (sometimes very beautiful, sometimes not)
// 2 = puts B/W ellipses for isovalues
// (synthetic, but not very beautiful)




////////////////////////////////////////////////////////////////////////////
//Do not modify the following variables (they are used to fit the data in the image,
//compute images parameters, define the dichotomic search for ct, etc.)
////////////////////////////////////////////////////////////////////////////

#define VRAIMENT_PAS_GRAND_CHOSE 0.00000001
#define TOUPITI 1.0e-30
#define BCP 100000000
#define TAILLE_POINTS 3
#define P 3
#define Compute_Neighbors 0
#define PI 3.141592654

double LIMIT_PERCENTAGE = 0.0000000001;
double C_MAX = 0.0005; 
double MIN_C = 0.0;
int NORMALISATION_CONST = 1;

double PERC_SIGNIFICATIF = 0.01; 
double EPSILON_KHM = 0.01; 
double EPSILON_LOG_KHM = 100; 
int P_KHM = 2; 

double EPSILON_FCM = 0.001;
int M_FCM = 2; 

double CONST_MOY_LIMITE=500; 
double CONST_SIGMA_LIMITE=100; 
int NEIGHBORS=15;
int NO_NEG_C=0;
int DISTANCE=0;

int STRETCH = 1;
int COORD_MAX = 300;

double NIVELLAGE;

int SHOW_WEIGHTS=0;
int SHOW_CONVEX_HULL=1;

int STEP_FOR_SAVING_IMG=1;
int INITIALISATION_FORGY=1;

int C_UNDEFINED;

static char BUF[12];

//#include"additional.h"
#include"cluster.h"
#include <stdlib.h>
#include <stdio.h>
#include <tiff.h>
#include <tiffio.h>
#include <math.h>

BOOL AFFICHE=FAUX;

couleur* GRIS_CLAIR=new couleur(220, 220, 220, 255);
couleur* VERT_MOYEN=new couleur(200, 255, 200, 255);

double puissance(double v, int e){
  double ret=1;
  int i;
  for (i=0;i<e;i++) ret*=v;
  return ret;
}

int tronque(double v){
  return (int) trunc(v);
}

double p(double d){
  //On suit exactement Hammerly + Elkan
  //retourne d_max^p
  double r, dp;
  dp = (d < EPSILON_KHM) ? EPSILON_KHM : d;
  r = puissance(dp,P_KHM);
  return r;
}

double ppd(double d){
  //On suit exactement Hammerly + Elkan
  //retourne d_max^{p+2}
  double r, dp;
  dp = (d < EPSILON_KHM) ? EPSILON_KHM : d;
  r = puissance(dp,P_KHM);
  r = r * dp * dp;
  return r;
}

/**************************************************************
* classe imageC
* implementation de la classe gerant les images de clustering
*****/

/*****
 * imageC::imageC(const char *file, int width, int length)
 * constructeur de la classe
 *****/
imageC::imageC(const char *file, int width, int length)
{
  filename=file;
  Wi=width;
  Le=length;
  bm= new (pixel*) [sizeof(pixel*)*Wi*Le];
  int i,j;
  for (i = 0; i < Le; i++) 
   {   	
     for (j = 0; j < Wi; j++) 
       {
	 bm[i*Wi+j]=new pixel;
	 bm[i*Wi+j]->r=255;
	 bm[i*Wi+j]->g=255;
	 bm[i*Wi+j]->b=255;
	 bm[i*Wi+j]->a=0;
       }
   }
}

/*****
* imageC::~imageC(void)
*****/
imageC::~imageC(void)
{
  int i,j;
  for (i = 0; i < Le; i++){
    for (j = 0; j < Wi; j++) 
       {
	 delete bm[i*Wi+j];
       }
  }
  delete [] bm;
}

/*****
 * void imageC::sauvegarde(void) 
 * methode de sauvegarde du tableau de l'image dans un TIFF RGBA 8 bits
 *****/
void imageC::sauvegarde(void) 
{
   unsigned int *data;
   data = new (unsigned int) [sizeof(unsigned int)*Wi*Le];
   //data = (unsigned int*) (malloc(sizeof(unsigned int)*Wi*Le));
   int pixelcolorR, pixelcolorG, pixelcolorB, pixelcolorAlpha;

   for (int i = 0; i < Le; i++) 
     {	
       for (int j = 0; j < Wi; j++) 
	 {   
	   pixelcolorR = bm[i*Wi+j]->r;
	   pixelcolorG = bm[i*Wi+j]->g;
	   pixelcolorB = bm[i*Wi+j]->b;
	   pixelcolorAlpha = bm[i*Wi+j]->a;
	   
	   data[i*Wi+j] = pixelcolorR
	     + (pixelcolorG << 8)
	     + (pixelcolorB << 16)
	     + (pixelcolorAlpha << 24);
	 }
     }

   //on renverse l'image
   int tmp;
   for (int i = 0; i < (Le/2); i++)
     {
       for (int j = 0; j < Wi; j++) 
	 {
	   tmp=data[(Le-1-i)*Wi+j];
	   data[(Le-1-i)*Wi+j]=data[i*Wi+j];
	   data[i*Wi+j]=tmp;
	 }
     }

   if (saveRGBtiff(filename, "", "", data, PHOTOMETRIC_RGB, Wi, Le) == 0)
	{	printf ("erreur readtiffgl"); }

   delete [] data;
}

/*****
 * int imageC::saveRGBtiff(const char *filename, const char *soft, const char *desc,
 * unsigned int *data, short photometric, int width, int height)
 *
 * annexe tiff
 *****/
int imageC::saveRGBtiff(const char *filename, const char *soft, const char *desc,
                   unsigned int *data, short photometric, int width, int height)
{
  unsigned int ori=1;
  TIFF *tif;
  int row;
  unsigned int *dummy;

  if((filename==NULL) || !strcmp(filename,"")) return 0;

  if(data==NULL){  /* if no output image given */
    fprintf(stderr,"sauvegarde image::pas d'image de sortie / no output image !\n");
    return 0;
  }

  tif = TIFFOpen(filename, "w");
  TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, width);
  TIFFSetField(tif, TIFFTAG_IMAGELENGTH, height);
  TIFFSetField(tif, TIFFTAG_COMPRESSION,COMPRESSION_NONE );
  TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
  TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 4);
  TIFFSetField(tif, TIFFTAG_ORIENTATION, ori);
  //TIFFSetField(tif, TIFFTAG_COLORMAP, ColorMap);
  TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, photometric);
  TIFFSetField(tif, TIFFTAG_ARTIST, "Nielsen and Nock");
  if(desc==NULL){             
    if(desc==NULL) desc = ""; 
    else                      
      TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, desc);
  }
  else
    TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, desc);
  TIFFSetField(tif, TIFFTAG_SOFTWARE, soft);

  dummy = data;
  for (row = 0; row < height; row++) {
    TIFFWriteScanline(tif, dummy, row, 0);
    dummy += 1*width;
  }

  TIFFClose(tif);

  return 1;
}

/**************************************************************
* classe tabPoints
* implementation de la classe gerant les tableaux de points
*****/

/*****
 * tabPoints::tabPoints(int dim, int gen, int bal, int n, int k)
 * constructeur de la classe
 *****/
tabPoints::tabPoints(int dim, int gen, int bal, int n, int k)
{
  int i;

  NP=n;
  KC_Theorique=k;
  dimension=dim;
  generation=gen;
  balance = bal;
  NN=NEIGHBORS;

  if (generation==0) CONST_SIGMA_LIMITE=100;
  else CONST_SIGMA_LIMITE=40;

  coordsPoints=new (int) [dimension*NP];
  maxC=new (int) [dimension];
  minC=new (int) [dimension];
  Nearest_Neighbor=new (int) [NN*NP];
  averageCoordCentres=new (int) [dimension*KC_Theorique];
  echantillonnage=new (double) [KC_Theorique];
  clusterTheorique=new (int) [NP];
  sigmaClusters=new (double) [KC_Theorique];

  for (i=0;i<dimension*NP;i++) coordsPoints[i]=0;

  for (i=0;i<dimension;i++) maxC[i]=0;
  for (i=0;i<dimension;i++) minC[i]=0;

  for (i=0;i<NN*NP;i++) Nearest_Neighbor[i]=0;
  for (i=0;i<dimension*KC_Theorique;i++) averageCoordCentres[i]=0;
  for (i=0;i<KC_Theorique;i++) sigmaClusters[i]=0;
  for (i=0;i<NP;i++) clusterTheorique[i]=0;

  if (bal==0){
    for (i=0;i<KC_Theorique;i++){
      echantillonnage[i]=1/( (double) KC_Theorique);
    }
  }else if (bal == 1){
    double valrand, summ=0;
    for (i=0;i<KC_Theorique;i++){
      valrand= (double) (abs(1 + rand()%1000));
      echantillonnage[i]=valrand;
      summ+=valrand;
    }
    for (i=0;i<KC_Theorique;i++)
      echantillonnage[i]/=summ;
  }else if (bal == 2){
    double reste=0;
    double val=0.5;
    for (i=0;i<KC_Theorique-1;i++){
      echantillonnage[i]=val;
      reste+=val;
    val /= (double) 2;
    }
    echantillonnage[KC_Theorique-1]=1-reste;
  }

  //for (i=0;i<KC_Theorique;i++) echantillonnage[i]=(double) 1/(double) KC_Theorique;
}

/*****
 * tabPoints::~tabPoints(void)
 * constructeur de la classe
 *****/
tabPoints::~tabPoints(void)
{
  delete [] coordsPoints;
  delete [] maxC;
  delete [] minC;
  delete [] Nearest_Neighbor;
  delete [] averageCoordCentres;
  delete [] echantillonnage;
  delete [] clusterTheorique;
}

/*****
 * tabPoints::GaussianSigma(double ave, double sigma)
 * tirage selon N(ave,sigma)
 *****/
double tabPoints::GaussianSigma(double ave, double sigma)
{
  static int iset=0;
  static double gset;
  double fac,rsq,v1,v2;

  if (iset == 0) {
    do {
      v1=2.0*((double)rand()/(double)RAND_MAX)-1.0;
      v2=2.0*((double)rand()/(double)RAND_MAX)-1.0;
      rsq=v1*v1+v2*v2;
    } while (rsq >= 1.0 || rsq==0.0);
    fac=sqrt(-2.0*log(rsq)/rsq);
    gset=v1*fac;
    iset=1;
    return ave+(v2*fac*sigma);
  } else {
    iset =0;
    return ave+(gset*sigma);
  }
}

/*****
 * void tabPoints::generate(BOOL Harmonic)
 * generation aleatoire des points
 * Harmonic = BOOLEEN : si VRAI, on divise tout par NORMALISATION_CONST (moyennes et ecart-types)
 *****/
void tabPoints::generate(BOOL Harmonic)
{
  generateAveSigma(Harmonic);
  if ( (generation==0) || (generation==3) ) generateGaussianClusters();
  else if (generation==1) generateRingGaussianClusters();
  else if (generation==2) generateGridGaussianClusters();
  else if (generation==4) generateCubicClusters();

  if (STRETCH == 1) stretchClusters();

  if (Compute_Neighbors==1) initNN();
}

/*****
 * void tabPoints::generateAveSigma(BOOL Harmonic)
 * generation aleatoire des esperances et sigma
 * Harmonic = BOOLEEN : si VRAI, on divise tout par NORMALISATION_CONST (moyennes et ecart-types)
 *****/
void tabPoints::generateAveSigma(BOOL Harmonic)
{
  int i,j;
  double v;

  for (i=0;i<KC_Theorique;i++){
    sigmaClusters[i]=((double)rand()/(double)RAND_MAX)*(double) (CONST_SIGMA_LIMITE/2);
    if (generation==3) sigmaClusters[i]=((double)rand()/(double)RAND_MAX)*(double) 20;
    //if (generation==1) sigmaClusters[i]/=10;

    for (j=0;j<dimension;j++){
      if (dimension<20){
	v=((double)rand()/(double)RAND_MAX)*CONST_MOY_LIMITE;
      }
      else{
	v=((double)rand()/(double)RAND_MAX)*(CONST_MOY_LIMITE/10);
      }
      putCoordCentre(i,j,tronque(v));
    }
  }

  if (Harmonic == VRAI){
    for (i=0;i<KC_Theorique;i++){
      sigmaClusters[i]/=(double) NORMALISATION_CONST;
      for (j=0;j<dimension;j++){
	putCoordCentre(i,j,(getCoordCentre(i,j)/NORMALISATION_CONST));
      }
    }
  }

  
  for (i=0;i<KC_Theorique;i++){
    if (sigmaClusters[i]<1) sigmaClusters[i]=1;
  }
}

/****
 * void tabPoints::stretchClusters(void)
 * reduit les coordonnees pour que les points tiennent dans un hypercube
 * de cote COORD_MAX
 *****/
void tabPoints::stretchClusters(void)
{
  int coordmin, coordmax, cc, ccm;
  double ccd, ccn;
  int i,j;
  int* cm = new (int) [dimension];
  for (j=0;j<dimension;j++) cm[j]=0;

  BOOL FIRSTMIN=VRAI;
  BOOL FIRSTMAX=VRAI;
  BOOL FIRSTTAB=VRAI;

  for (i=0;i<KC_Theorique;i++){
    for (j=0;j<dimension;j++){
      cc = getCoordCentre(i,j);
      if ( (FIRSTMIN == VRAI) || (cc < coordmin) ){
	coordmin = cc;
	FIRSTMIN = FAUX;
      }
      if ( (FIRSTMAX == VRAI) || (cc > coordmax) ){
	coordmax = cc;
	FIRSTMAX = FAUX;
      }
      if ( (FIRSTTAB == VRAI) || (cc < cm[j]) ){
	FIRSTTAB = FAUX;
	cm[j] = cc;
      }
    }
  }

  for (i=0;i<NP;i++){
    for (j=0;j<dimension;j++){
      cc = getCoordPoints(i,j);
      if ( (FIRSTMIN == VRAI) || (cc < coordmin) ){
	coordmin = cc;
	FIRSTMIN = FAUX;
      }
      if ( (FIRSTMAX == VRAI) || (cc > coordmax) ){
	coordmax = cc;
	FIRSTMAX = FAUX;
      }
      if ( (FIRSTTAB == VRAI) || (cc < cm[j]) ){
	FIRSTTAB = FAUX;
	cm[j] = cc;
      }
    }
  }

  for (i=0;i<KC_Theorique;i++){
    for (j=0;j<dimension;j++){
      ccn = ( (double) (getCoordCentre(i,j) - cm[j]) ) * COORD_MAX;
      ccd = (double) coordmax - coordmin;
      ccm = tronque ( (ccn / ccd) );
      putCoordCentre(i,j,ccm);
    }
  }

  for (j=0;j<dimension;j++){
    FIRSTMIN = VRAI;
    FIRSTMAX = VRAI;
    for (i=0;i<NP;i++){
      ccn = (double) (getCoordPoints(i,j) - cm[j])* COORD_MAX;
      ccd = (double) coordmax - coordmin;
      ccm = tronque ( (ccn / ccd) );
      if ( (FIRSTMIN == VRAI) || (ccm<minC[j]) ){
	FIRSTMIN = FAUX;
	minC[j] = ccm;
      }
      if ( (FIRSTMAX == VRAI) || (ccm>maxC[j]) ){
	FIRSTMAX = FAUX;
	maxC[j] = ccm;
      }
      putCoordPoints(i,j,ccm);
    }
  }

  //printf("%d %d\n",coordmin, coordmax);
  //affiche();

  printf("(dataset prepared)\n");

  delete [] cm;
}

/*****
 * void tabPoints::generateGaussianClusters(void)
 * genere des clusters gaussiens en ND
 *****/
void tabPoints::generateGaussianClusters(void)
{
  int i,j,k,nc;
  int minvalue = 10000;
  double valal,fmin,fmax,restnorm,nr;
  double* vecteur=new double[dimension];
  double* cp=new double[dimension];

  for (i=0;i<NP;i++){
    valal=(double)rand()/(double)RAND_MAX;
    if (valal==0) valal=echantillonnage[0]/2;
    nc=0;
    fmin=0;
    fmax=echantillonnage[0];
    while( ! ( (valal<=fmax) && (valal>fmin) ) ){
      fmin+=echantillonnage[nc];
      nc++;
      fmax+=echantillonnage[nc];
    }
    clusterTheorique[i]=nc;

    restnorm=0;
    for (j=0;j<dimension;j++){
      cp[j] = (double) (-(minvalue/2) + abs(rand()%minvalue));
      restnorm+=(cp[j]*cp[j]);
    }
    restnorm = sqrt(restnorm);
    ASSERT(restnorm>0);

    for (j=0;j<dimension;j++){
      cp[j]/=restnorm;
    }
    nr=GaussianSigma(0,sigmaClusters[nc]);
    for (j=0;j<dimension;j++){
      putCoordPoints(i,j,tronque( ((double) getCoordCentre(nc,j)) +(cp[j]*nr)));
      if ( (i==0) || (getCoordPoints(i,j)>maxC[j]) ) maxC[j]=getCoordPoints(i,j);
      if ( (i==0) || (getCoordPoints(i,j)<minC[j]) ) minC[j]=getCoordPoints(i,j);
    }
  }
}


/*****
 * void tabPoints::generateCubicClusters(void)
 * genere des clusters Cubiques en ND, dont chaque composante
 * des points est uniformement choisie entre M-2S et M+2S
 *****/
void tabPoints::generateCubicClusters(void)
{
  int i,j,k,nc,icc;
  double valal,fmin,fmax,cc;

  for (i=0;i<NP;i++){
    valal=(double)rand()/(double)RAND_MAX;
    if (valal==0) valal=echantillonnage[0]/2;
    nc=0;
    fmin=0;
    fmax=echantillonnage[0];
    while( ! ( (valal<=fmax) && (valal>fmin) ) ){
      fmin+=echantillonnage[nc];
      nc++;
      fmax+=echantillonnage[nc];
    }
    clusterTheorique[i]=nc;

    for (j=0;j<dimension;j++){
      icc = (int) sigmaClusters[nc];
      cc = ((double) getCoordCentre(nc,j)) + (double) (3*icc) - ( (double) 6*(rand()%icc));
      putCoordPoints(i,j,tronque(cc));
      if ( ( (i==0) && (j==0) ) || (getCoordPoints(i,j)>maxC[j]) ) maxC[j]=getCoordPoints(i,j);
      if ( ( (i==0) && (j==0) ) || (getCoordPoints(i,j)<minC[j]) ) minC[j]=getCoordPoints(i,j);
    }
  }
}


/*****
 * void tabPoints::generateRingGaussianClusters(void)
 * genere des clusters gaussiens sur des anneaux en ND
 *****/
void tabPoints::generateRingGaussianClusters(void)
{
  int i,j,k,nc;
  int minvalue = 10000;
  double valal,fmin,fmax,restnorm,nr;
  double* vecteur=new double[dimension];
  double* cp=new double[dimension];

  double* ring=new double[KC_Theorique];
  for (i=0;i<KC_Theorique;i++){
    ring[i]=(double) (rand()%((int) CONST_MOY_LIMITE/10));
  }

  for (i=0;i<NP;i++){
    valal=(double)rand()/(double)RAND_MAX;
    if (valal==0) valal=echantillonnage[0]/2;
    nc=0;
    fmin=0;
    fmax=echantillonnage[0];
    while( ! ( (valal<=fmax) && (valal>fmin) ) ){
      fmin+=echantillonnage[nc];
      nc++;
      fmax+=echantillonnage[nc];
    }
    clusterTheorique[i]=nc;


    restnorm=0;
    for (j=0;j<dimension;j++){
      cp[j] = (double) (-(minvalue/2) + abs(rand()%minvalue));
      restnorm+=(cp[j]*cp[j]);
    }
    restnorm = sqrt(restnorm);
    ASSERT(restnorm>0);

    for (j=0;j<dimension;j++){
      cp[j]/=restnorm;
    }
    nr=GaussianSigma(0,sigmaClusters[nc]);
    for (j=0;j<dimension;j++){
      putCoordPoints(i,j,tronque( ((double) getCoordCentre(nc,j)) +(cp[j]*(nr+5*ring[nc]))));
      if ( (i==0) || (getCoordPoints(i,j)>maxC[j]) ) maxC[j]=getCoordPoints(i,j);
      if ( (i==0) || (getCoordPoints(i,j)<minC[j]) ) minC[j]=getCoordPoints(i,j);
    }
  }

  delete [] ring;
}

/*****
 * int tabPoints::puiss(int a, int b)
 * renvoie un int a^b
 *****/
int tabPoints::puiss(int a, int b)
{
  int i;
  int r=a;
  for (i=0;i<b-1;i++) r*=a;
  return r;
}

/*****
 * int tabPoints::plus_pt_int(int a, int b)
 * renvoie le plus petit entier c tel que c^a >= b
 *****/
int tabPoints::plus_pt_int(int a, int b)
{
  int r=1;
  while (puiss(r,a)<b) r++;
  return r;
}

/*****
 * void tabPoints::generateGridGaussianClusters(void)
 * genere des clusters gaussiens en 2D sur une grille
 * (jeu de donnees de type BIRCH)
 *****/
void tabPoints::generateGridGaussianClusters(void)
{
  if (dimension != 2)
    printf("BIRCH dataset MUST be chosen with dimension = 2, please run again\n");
  assert(dimension == 2);

  int i,j,k,nc;
  double valal,fmin,fmax,restnorm,nr;
  double* vecteur=new double[dimension];
  double* cp=new double[dimension];

  assert(dimension==2);

  int kdim=plus_pt_int(dimension,KC_Theorique);
  //recalcul des moyennes et ecarts-types
  double ec=CONST_MOY_LIMITE/(double) kdim;
  double sis=ec/4;
  int nuf=0;
  for (i=0;i<kdim;i++){
    for (j=0;j<kdim;j++){
      if (nuf<KC_Theorique){
	putCoordCentre(((i*kdim)+j),0,tronque((ec/2)+(i*ec)));
	putCoordCentre(((i*kdim)+j),1,tronque((ec/2)+(j*ec)));
      }
      nuf++;
    }
  }
  for (i=0;i<KC_Theorique;i++)sigmaClusters[i]=sis;
  

  for (i=0;i<NP;i++){
    valal=(double)rand()/(double)RAND_MAX;
    if (valal==0) valal=echantillonnage[0]/2;
    nc=0;
    fmin=0;
    fmax=echantillonnage[0];
    while( ! ( (valal<=fmax) && (valal>fmin) ) ){
      fmin+=echantillonnage[nc];
      nc++;
      fmax+=echantillonnage[nc];
    }
    clusterTheorique[i]=nc;

    for (j=0;j<dimension;j++) cp[j]=2;
    restnorm=1;
    for (j=0;j<dimension;j++){
      do{
	k=abs(rand()%dimension);
      }while(cp[k]!=2);
      if (restnorm==0) cp[k]=0;
      else if (j==dimension-1){
	cp[k]=( (rand()%2==0) ?  -sqrt(restnorm) : sqrt(restnorm) );
      } else {
	cp[k]=(2*((double)rand()/(double)RAND_MAX)-1)*sqrt(restnorm);
      }
      assert( (cp[k]>=-sqrt(restnorm) - VRAIMENT_PAS_GRAND_CHOSE) 
	      && (cp[k]<=sqrt(restnorm) + VRAIMENT_PAS_GRAND_CHOSE) );
      restnorm-=(cp[k]*cp[k]);
    }
    assert( (restnorm<=VRAIMENT_PAS_GRAND_CHOSE) && (restnorm>=-VRAIMENT_PAS_GRAND_CHOSE) );
    nr=GaussianSigma(0,sigmaClusters[nc]);

    for (j=0;j<dimension;j++){
      putCoordPoints(i,j,tronque((double) getCoordCentre(nc,j)+(cp[j]*nr)));
      if ( (i==0) || (getCoordPoints(i,j)>maxC[j]) ) maxC[j]=getCoordPoints(i,j);
      if ( (i==0) || (getCoordPoints(i,j)<minC[j]) ) minC[j]=getCoordPoints(i,j);
    }
  }
}

/*****
 * void tabPoints::initNN(void)
 * initialise les NN
 *****/
void tabPoints::initNN(void)
{
  int i,j,k,l;
  double distmin, dist;
  int * Vect=new int[NN];
  BOOL trouve;
  for (i=0;i<NP;i++){
    for (j=0;j<NN;j++) Vect[j]=-1;
    for (k=0;k<NN;k++){
      distmin=-1;
      for (j=0;j<NP;j++){
	trouve=FAUX;
	if (j==i) trouve=VRAI;
	l=0;
	do{
	  if (Vect[l]==j) trouve=VRAI;
	  l++;
	}while( (l<NN) && (trouve==FAUX) );

	if (!trouve){
	  dist=distance(i,j);
	  if ( (distmin==-1) || (dist<distmin) ){
	    Vect[k]=j;
	    distmin=dist;
	  }
	}
      }
    }

    for (j=0;j<NN;j++){
      putNeighbor(i,j,Vect[j]);
    }
  }
  delete [] Vect;
}

/*****
 * void tabPoints::affiche(void)
 * affiche les coords des points et leur cluster theorique
 *****/
void tabPoints::affiche(void)
{
  int i,j;
  printf("NP=%d, KC_Theorique=%d, dimension=%d\n",NP,KC_Theorique,dimension);
  for (i=0;i<NP;i++){
    printf("pt #%d:\t ",i);
    for (j=0;j<dimension;j++){
      printf("[%d:%d] ",j,getCoordPoints(i,j));
    }
    printf(" <-%d\n", clusterTheorique[i]);
  }

  for (i=0;i<dimension;i++){
    printf("[%d]: min=%d, max=%d\n",i,minC[i],maxC[i]);
  }
}

/*****
 * void tabPoints::claque_im(imageC* im, BOOL colors)
 * place les points dans l'image im selon les deux premiere coords
 * si colors==VRAI, on met les couleurs theoriques
 *****/
void tabPoints::claque_im(imageC* im, BOOL colors)
{
  int i,j,r,g,b,a,x,y;
  double redv, greenv, bluev;
  double ratioX, ratioY;

  //dessin du graphe des NN
  if ( (dimension>1) && (Compute_Neighbors==1) ){
    for (i=0;i<NP;i++){
      for (j=0;j<NN;j++){
	int x1, x2, y1, y2;

	ratioX=((double) (im->vWi()))/((double) (maxC[0]-minC[0]));
	ratioY=((double) (im->vLe()))/((double) (maxC[1]-minC[1]));

	x1=(int) (((double) (getCoordPoints(i,0)-minC[0]))*ratioX);
	x2=(int) (((double) (getCoordPoints(getNeighbor(i,j),0)-minC[0]))*ratioX);

	y1=(int) (((double) (getCoordPoints(i,1)-minC[1]))*ratioY);
	y2=(int) (((double) (getCoordPoints(getNeighbor(i,j),1)-minC[1]))*ratioY);

	im->ligne(x1, y1, x2, y2, 
		  GRIS_CLAIR->getR(), 
		  GRIS_CLAIR->getG(), 
		  GRIS_CLAIR->getB(), 
		  GRIS_CLAIR->getA(), 1);
      }
    }
  }

  for (i=0;i<NP;i++){
    if (colors){
      redv=((double) clusterTheorique[i] + 1)/( (double) KC_Theorique);
      greenv=( (redv<=0.5) ?  (2*redv) : (2*(1-redv)) );
      bluev=1-redv;
    }else{
      redv=1;
      greenv=1;
      bluev=1;
    }
    r=(int) (redv*255);
    g=(int) (greenv*255);
    b=(int) (bluev*255);
    a=255;

    ratioX=((double) (im->vWi()))/((double) (maxC[0]-minC[0]));
    x=(int) (((double) (getCoordPoints(i,0)-minC[0]))*ratioX);

    if (dimension>1){
      ratioY=((double) (im->vLe()))/((double) (maxC[1]-minC[1]));
      y=(int) (((double) (getCoordPoints(i,1)-minC[1]))*ratioY);
    }else{
      y=(im->vLe())/2;
    }

    im->point(x,y,r,g,b,a,TAILLE_POINTS);
  }
}


/**************************************************************
* classe clusters
* implementation de la classe gerant les algos de clustering
*****/

/*****
 * clusters::clusters(tabPoints* tab, int K, char* nf, int t_i, int S, int W, int C, BOOL INIT)
 * constructeur de la classe
 *****/
clusters::clusters(tabPoints* tab, int K, char* nf, int t_i, int S, int W, int C, int I, BOOL INIT)
{
  smethod = new (char) [255];

  SHOW_WEIGHTS=W;
  SHOW_CONVEX_HULL=C;
  STEP_FOR_SAVING_IMG=S;
  INITIALISATION_FORGY=I;

  int i;

  sfile=nf;
  pointSet=tab;
  NP=pointSet->vNP();
  dimension=pointSet->vdim();
  KC=K;
  t_im=t_i;

  weights=new (double) [NP];
  maxWeight=(double) 1/(double) NP;
  minWeight=(double) 1/(double) NP;
  membershipClusters=new (double) [KC*NP];
  centersCoordFromClusters=new (double) [KC*dimension];
  clusterValide=new (BOOL) [KC];

  varcovar = new (matrice*) [KC];
  inverseVarcovar = new (matrice*) [KC];
  det_man = new (double) [KC];
  det_exp = new (int) [KC];
  mixingProp = new (double) [KC];

  Log_Loss_Prec=new (double) [NP];
  Log_Loss_Cour=new (double) [NP];
  for (i=0;i<NP;i++) weights[i]=(double) 1/(double) NP;
  for (i=0;i<NP;i++) Log_Loss_Prec[i]=0;
  for (i=0;i<NP;i++) Log_Loss_Cour[i]=0;
  for (i=0;i<KC;i++) clusterValide[i] = VRAI;
  for (i=0;i<KC;i++){
    mixingProp[i] = 1 / (double) KC;
    det_man[i]=0;
    det_exp[i]=0;
  }

  for (i=0;i<KC;i++) varcovar[i] = new matrice(dimension);
  for (i=0;i<KC;i++) inverseVarcovar[i] = new matrice(dimension);

  L_KMN=0;
  Min_Prop = 1;
  Log_Prop = 1;
  Prop_Missed_Fin_Algo = 1;
  Max_Loss = -1;
  // attention a les initialiser proprement !

  if (INIT == VRAI)
    if (INITIALISATION_FORGY == 1)
      init_clusters_centers_Forgy();
    else
      init_clusters_centers_Random();
}

/*****
 * void clusters::reinitialise_clustering(clusters* cap)
 * initialise le clustering a partir d'un autre clustering
 *****/
void clusters::reinitialise_clustering(clusters* cap)
{
  copie_cluster_centers(cap);
  //affecte_points_clusters();
  //Compute_L_KMN();
  //Compute_Variabilite_Courante();

  //sauvegarde_im(-1);
}

/*****
 * clusters::~clusters(void)
 * destructeur de la classe
 *****/
clusters::~clusters(void)
{
  delete [] smethod;
  delete [] weights;
  delete [] membershipClusters;
  delete [] centersCoordFromClusters;

  delete [] Log_Loss_Prec;
  delete [] Log_Loss_Cour;

  int i;
  for (i=0;i<KC;i++)
    delete [] varcovar[i];
  delete [] varcovar;

  for (i=0;i<KC;i++)
    delete [] inverseVarcovar[i];
  delete [] inverseVarcovar;

  delete [] mixingProp;
  delete [] det_man;
  delete [] det_exp;
}


/*****
 * void clusters::copie_cluster_centers(clusters* cap)
 * copie les centres de clusters choisis pour cap dans le receveur
 *****/
void clusters::copie_cluster_centers(clusters* cap)
{
  int i,k;
  for (i=0;i<KC;i++){
    putClusterValide(i,cap->getClusterValide(i));
    for (k=0;k<dimension;k++) putCoord(centersCoordFromClusters,i,k,cap->getCoordCenter(i,k));
  }
}


/*****
 * void clusters::claque_im(imageC* im)
 * place les clusters empiriques dans l'image im selon les deux premiere coords
 *****/
void clusters::claque_im(imageC* im)
{
  int i,j,r,g,b,a,x,y;
  double redv, greenv, bluev;
  double ratioX, ratioY;
  
  double param;


  if (DISTANCE==1){
    double xbef, ybef;
    int k,l;
    double deno;
    double qred, qgre, qblu;
    double* npc=new (double) [KC];

    for (i=0;i<im->vLe();i++){
      for (j=0;j<im->vWi();j++){
	ratioX=((double) (im->vWi()))/( ((double) (pointSet->vMax(0)-pointSet->vMin(0))) );
	ratioY=((double) (im->vLe()))/( ((double) (pointSet->vMax(1)-pointSet->vMin(1))) );

	for (k=0;k<KC;k++) npc[k]=0;
	deno=0;
	for (k=0;k<KC;k++){
	  if (getClusterValide(k)==VRAI){
	    xbef=getCoord(centersCoordFromClusters,k,0);
	    x=(int) ((xbef- ((double) pointSet->vMin(0)) )*ratioX);
	    if (dimension>1){
	      ybef=getCoord(centersCoordFromClusters,k,1);
	      y=(int) ((ybef- ((double) pointSet->vMin(1)) )*ratioY);
	    }else{
	      y=(im->vLe())/2;
	    }
	    npc[k]=(i-x)*(i-x)+(j-y)*(j-y);
	    npc[k]=sqrt(npc[k]);
	    npc[k]=puissance(npc[k],P+2);
	    npc[k]=1/npc[k];
	    deno+=npc[k];

	    //Si Harmonic Clustering, on trace la fonction de cout soft en projetant les points
	    //dans le plan (x,y).
	  }
	}
	qred=0;
	qgre=0;
	qblu=0;

	for (k=0;k<KC;k++){
	  //redv=max( (0) , ( ((double) 2/KC)*k - 1 ) );
	  //greenv=( (k<=(KC/2)) ? (((double) 2/KC)*k) : (2- (((double) 2/KC)*k) ) );
	  //bluev=max( (0) , ( (1-(((double) 2/KC)*k)) ) );

	  redv=((double) k + 1)/( (double) KC);
	  greenv=( (redv<=0.5) ?  (2*redv) : (2*(1-redv)) );
	  bluev=1-redv;

	  qred+=redv*npc[k]/deno;
	  qgre+=greenv*npc[k]/deno;
	  qblu+=bluev*npc[k]/deno;
	}

	r=(int) (qred*255);
	g=(int) (qgre*255);
	b=(int) (qblu*255);
	a=255;
    
	im->point(i,j,r,g,b,a,1);
      }
    }

    delete [] npc;
  }


  if (DISTANCE==3){
    double xbef, ybef, anum, aden, aA, aB, apuiss, acontribcour;
    int k,l,m;
    double deno;
    double qred, qgre, qblu;
    double* npc=new (double) [KC];

    for (i=0;i<im->vLe();i++){
      for (j=0;j<im->vWi();j++){
	ratioX=((double) (im->vWi()))/( ((double) (pointSet->vMax(0)-pointSet->vMin(0))) );
	ratioY=((double) (im->vLe()))/( ((double) (pointSet->vMax(1)-pointSet->vMin(1))) );

	for (k=0;k<KC;k++) npc[k]=0;
	deno=0;
	for (k=0;k<KC;k++){
	  if (getClusterValide(k)==VRAI){
	    xbef=getCoord(centersCoordFromClusters,k,0);
	    x=(int) ((xbef- ((double) pointSet->vMin(0)) )*ratioX);
	    if (dimension>1){
	      ybef=getCoord(centersCoordFromClusters,k,1);
	      y=(int) ((ybef- ((double) pointSet->vMin(1)) )*ratioY);
	    }else{
	      y=(im->vLe())/2;
	    }

	    anum = sqrt((double) (i-x)*(i-x)+(j-y)*(j-y));
	    for (m=0;m<KC;m++){
	      if (getClusterValide(m)==VRAI){
		xbef=getCoord(centersCoordFromClusters,m,0);
		x=(int) ((xbef- ((double) pointSet->vMin(0)) )*ratioX);
		if (dimension>1){
		  ybef=getCoord(centersCoordFromClusters,m,1);
		  y=(int) ((ybef- ((double) pointSet->vMin(1)) )*ratioY);
		}else{
		  y=(im->vLe())/2;
		}

		aden = sqrt((double) (i-x)*(i-x)+(j-y)*(j-y));
		if (aden < EPSILON_FCM)
		  aden = EPSILON_FCM;
		
		aA = anum / aden;
		aB = (double) 2 / (double) (M_FCM - 1);
		apuiss = pow(aA, aB);
		acontribcour += apuiss;
	      }
	    }
	    if (acontribcour < EPSILON_FCM)
	      acontribcour = EPSILON_FCM;
	    acontribcour = 1 / acontribcour;
	    npc[k] = acontribcour;

	    deno+=npc[k];

	    //Si Fuzzy C Means, on trace la fonction de cout soft en projetant les points
	    //dans le plan (x,y).
	  }
	}
	qred=0;
	qgre=0;
	qblu=0;

	for (k=0;k<KC;k++){
	  //redv=max( (0) , ( ((double) 2/KC)*k - 1 ) );
	  //greenv=( (k<=(KC/2)) ? (((double) 2/KC)*k) : (2- (((double) 2/KC)*k) ) );
	  //bluev=max( (0) , ( (1-(((double) 2/KC)*k)) ) );

	  redv=((double) k + 1)/( (double) KC);
	  greenv=( (redv<=0.5) ?  (2*redv) : (2*(1-redv)) );
	  bluev=1-redv;

	  qred+=redv*npc[k]/deno;
	  qgre+=greenv*npc[k]/deno;
	  qblu+=bluev*npc[k]/deno;
	}

	r=(int) (qred*255);
	g=(int) (qgre*255);
	b=(int) (qblu*255);
	a=255;
    
	im->point(i,j,r,g,b,a,1);
      }
    }

    delete [] npc;
  }

  double* redcol = new (double) [KC];
  double* greencol = new (double) [KC];
  double* bluecol = new (double) [KC];

  for (j=0;j<KC;j++){
    redcol[j] = (double) j / (double) (KC-1);
    greencol[j] = (j<KC/2) ? 2*redcol[j] : 2*(1-redcol[j]);
    bluecol[j] = 1 - redcol[j];
  }

  if (DISTANCE==2){
    double xbef, ybef;
    int k,l;
    double deno;
    double qred, qgre, qblu;
    double* npc=new (double) [KC];
    double dx, dy, s1, s2, s12, s21;
    double lim = (double) (im->vWi() / 20);
    int bordure;

    for (i=0;i<im->vLe();i++){
      for (j=0;j<im->vWi();j++){
	bordure = 0;
	ratioX=((double) (im->vWi()))/( ((double) (pointSet->vMax(0)-pointSet->vMin(0))) );
	ratioY=((double) (im->vLe()))/( ((double) (pointSet->vMax(1)-pointSet->vMin(1))) );

	for (k=0;k<KC;k++) npc[k]=0;
	deno=0;
	for (k=0;k<KC;k++){
	  if (getClusterValide(k)==VRAI){
	    xbef=getCoord(centersCoordFromClusters,k,0);
	    x=(int) ((xbef- ((double) pointSet->vMin(0)) )*ratioX);
	    if (dimension>1){
	      ybef=getCoord(centersCoordFromClusters,k,1);
	      y=(int) ((ybef- ((double) pointSet->vMin(1)) )*ratioY);
	    }else{
	      y=(im->vLe())/2;
	    }
	    dx = (i-x);
	    dy = (j-y);
	    s1 = inverseVarcovar[k]->getCoord(0,0);
	    s2 = inverseVarcovar[k]->getCoord(1,1);
	    s12 = inverseVarcovar[k]->getCoord(0,1);
	    s21 = inverseVarcovar[k]->getCoord(1,0);
	    
	    npc[k] = (s1*dx*dx) + ((s12+s21)*dx*dy) + (s2*dy*dy);

	    if ( (npc[k]<=lim) && (npc[k]>=lim*0.8) )
	      bordure = 1;

	    if (EM_PUT_ISOVALEURS == 1){
	      if (npc[k]<=lim) npc[k]=0;
	      else npc[k]=1;
	    }

	    deno+=npc[k];

	    //Si EM, on trace la fonction de cout soft en projetant les points
	    //dans le plan (x,y).
	  }
	}
	qred=0;
	qgre=0;
	qblu=0;

	for (k=0;k<KC;k++){
	  if (getClusterValide(k)==VRAI){
	    redv=redcol[k];
	    greenv=greencol[k];
	    bluev=bluecol[k];
	    
	    qred+=redv*npc[k]/deno;
	    qgre+=greenv*npc[k]/deno;
	    qblu+=bluev*npc[k]/deno;
	  }
	}

	if (EM_PUT_ISOVALEURS == 2){
	  if (bordure == 1){
	    for (k=0;k<KC;k++){
	      qred = 0;
	      qgre = 0;
	      qblu = 0;
	    }
	  }else{
	    for (k=0;k<KC;k++){
	      qred = 1;
	      qgre = 1;
	      qblu = 1;
	    }
	  }
	}

	r=(int) (qred*255);
	g=(int) (qgre*255);
	b=(int) (qblu*255);

	if (EM_PUT_ISOVALEURS == 1){
	  r = 64 + (3*r/4);
	  g = 64 + (3*g/4);
	  b = 64 + (3*b/4);
	}

	a=255;
    
	im->point(i,j,r,g,b,a,1);
      }
    }

    delete [] npc;
  }

  for (i=0;i<NP;i++){
    redv=0;
    for (j=0;j<KC;j++){
      redv += ( (double) j + 1 )*getMembership(j,i);
    }
    redv/=( (double) KC);
    greenv=( (redv<=0.5) ?  (2*redv) : (2*(1-redv)) );
    bluev=1-redv;

    if (DISTANCE==2){
      redv = 0;
      greenv = 0;
      bluev = 0;
      for (j=0;j<KC;j++){
	redv += redcol[j]*getMembership(j,i);
	greenv += greencol[j]*getMembership(j,i);
	bluev += bluecol[j]*getMembership(j,i);
      }
    }
    
    r=(int) (redv*255);
    g=(int) (greenv*255);
    b=(int) (bluev*255);
    a=255;
    
    if ( (DISTANCE==1) || ( (DISTANCE==2) && (EM_PUT_ISOVALEURS<2) ) || (DISTANCE==3)  ){
      r=0;
      g=0;
      b=0;
      a=0;
    }

    ratioX=((double) (im->vWi()))/( (double) (pointSet->vMax(0)-pointSet->vMin(0)) );
    x=(int) ( ((double) (pointSet->getCoordPoints(i,0)-pointSet->vMin(0))) *ratioX);
    
    if (dimension>1){
      ratioY=((double) (im->vLe()))/( (double) (pointSet->vMax(1)-pointSet->vMin(1)) );
      y=(int) ( ( (double) (pointSet->getCoordPoints(i,1)-pointSet->vMin(1)) )*ratioY);
    }else{
      y=(im->vLe())/2;
    }
    
    if (SHOW_WEIGHTS==1){
      if (maxWeight==minWeight) param=1;
      else param=(weights[i]-minWeight)/(maxWeight-minWeight);
      
      im->cercle(x,y,(int) (param*(im->vWi() / 20)),
		 VERT_MOYEN->getR(),
		 VERT_MOYEN->getG(),
		 VERT_MOYEN->getB(),
		 VERT_MOYEN->getA(),1);
    }

    im->point(x,y,r,g,b,a,TAILLE_POINTS);
  }
  

  if ( (dimension>1) && (SHOW_CONVEX_HULL==1) && (DISTANCE==0) ) put_convex_hulls(im);


  if (dimension<=300){
    for (i=0;i<KC;i++){
      if (getClusterValide(i)==VRAI){
	x=(int) (( (double) (getCoord(centersCoordFromClusters,i,0)-pointSet->vMin(0)) )*ratioX);
	if (dimension>1){
	  y=(int) (( (double) (getCoord(centersCoordFromClusters,i,1)-pointSet->vMin(1)) )*ratioY);
	}else{
	  y=(im->vLe())/2;
	}
	
	if (DISTANCE==0) im->point(x,y,0,0,0,0,TAILLE_POINTS*2);
	if ( (DISTANCE==1) || (DISTANCE==2) || (DISTANCE==3) ) im->point(x,y,255,255,255,0,TAILLE_POINTS*2);
      }
    }
  }

  delete [] redcol;
  delete [] bluecol;
  delete [] greencol;
}

/*****
 * void clusters::put_convex_hulls(imageC* im)
 * place les enveloppes convexes des clusters empiriques dans l'image im selon les deux premiere coords
 *****/
void clusters::put_convex_hulls(imageC* im)
{
  int i,j,k,te,r,g,b,a,x1,x2,y1,y2,test;
  int firstx, firsty;
  double redv, greenv, bluev;
  double ratioX, ratioY;

  int* cardK;
  cardK=new int[KC];

  double* Px, *Py;
  double* Hx, *Hy;

  BOOL SORTIE;

  ratioX=((double) (im->vWi()))/( (double) (pointSet->vMax(0)-pointSet->vMin(0)) );
  ratioY=((double) (im->vLe()))/( (double) (pointSet->vMax(1)-pointSet->vMin(1)) );

  for (i=0;i<KC;i++)cardK[i]=0;
  for (i=0;i<NP;i++)
    for (j=0;j<KC;j++)
      cardK[j]+=(getMembership(j,i)==1) ? 1 : 0;
  //cardK[clusterEmpirique[i]]++;

  for (i=0;i<KC;i++){
    //if (AFFICHE){
    //  printf("Cardk[%d]=%d\n ",i,cardK[i]);
    //  fflush(stdout);
    //}

    if (cardK[i]>0){

      redv=((double) i + 1)/( (double) KC);
      greenv=( (redv<=0.5) ?  (2*redv) : (2*(1-redv)) );
      bluev=1-redv;
      
      r=(int) (redv*255);
      g=(int) (greenv*255);
      b=(int) (bluev*255);
      a=255;
      
      Px=(double*) malloc(sizeof(double)*cardK[i]);
      Py=(double*) malloc(sizeof(double)*cardK[i]);
      Hx=(double*) malloc(sizeof(double)*cardK[i]);
      Hy=(double*) malloc(sizeof(double)*cardK[i]);
      
      //Px=new (double) [cardK[i]];
      //Py=new (double) [cardK[i]];
      //Hx=new (double) [cardK[i]];
      //Hy=new (double) [cardK[i]];
      
      k=0;
      for (j=0;j<NP;j++){
	if (getMembership(i,j)==1){
	  Px[k]=(double) pointSet->getCoordPoints(j,0);
	  Py[k]=(double) pointSet->getCoordPoints(j,1);
	  k++;
	}
      }
      sort(Px,Py,k);
      
      /*if (AFFICHE){
	printf("%d\n ",i);
	fflush(stdout);
	}*/
      
      te=chainHull_2D(Px,Py,k,Hx,Hy);
      
      SORTIE=FAUX;
      j=0;
      
      while (SORTIE==FAUX){ 
	
	x1=(int) ((Hx[j]- ((double) pointSet->vMin(0)) )*ratioX);
	x2=(int) ((Hx[j+1]- ((double) pointSet->vMin(0)) )*ratioX);
	
	y1=(int) ((Hy[j]- ((double) pointSet->vMin(1)) )*ratioY);
	y2=(int) ((Hy[j+1]- ((double) pointSet->vMin(1)) )*ratioY);
	
	if (j==0){
	  firstx=x1;
	  firsty=y1;
	}
	
	/*if (AFFICHE){
	  printf("--%d/%d (%d:%d)-(%d:%d) \n ",j,te,x1,y1,x2,y2);
	  fflush(stdout);
	  }*/
	
	im->ligne(x1,y1,x2,y2,r,g,b,a,1);
	
	/*if (AFFICHE){
	  printf("*-%d/%d (%d:%d)-(%d:%d) \n ",j,te,x1,y1,x2,y2);
	  fflush(stdout);
	  }*/
	
	j++;
	if ( (x2==firstx) && (y2==firsty) ) SORTIE=VRAI;
      }
      
      /*if (AFFICHE){
	printf("/-%d/%d (%d:%d)-(%d:%d) \n ",j,te,x1,y1,x2,y2);
	fflush(stdout);
	}*/
      
      /*if (AFFICHE){
	printf("%p, %p, %p, %p \n ",Px,Py,Hx,Hy);
	fflush(stdout);
	}*/
      
      free((void*) Hy);
      free((void*) Hx);
      free((void*) Py);
      free((void*) Px);
      Px=NULL;
      Py=NULL;
      Hx=NULL;
      Hy=NULL;


      /*if (AFFICHE){
	printf("%p, %p, %p, %p \n ",Px,Py,Hx,Hy);
	fflush(stdout);
	}*/
      
      //delete [] Px;
      //delete [] Py;
      //delete [] Hx;
      //delete [] Hy;
    }
  }

  delete [] cardK;

  /*if (AFFICHE){
	printf("*-sortie \n ");
	fflush(stdout);
	}*/
}

/*****
 * double clusters::isLeft(double p0x, double p0y, double p1x, double p1y, double p2x, double p2y)
 * return >0 for P2 left of the line through P0 and P1
 * =0 for P2 on the line
 *            <0 for P2 right of the line
 *****/
double clusters::isLeft(double p0x,
		   double p0y,
		   double p1x,
		   double p1y,
		   double p2x,
		   double p2y)
{
    return (p1x - p0x)*(p2y - p0y) - (p2x - p0x)*(p1y - p0y);
}

/*****
 * void clusters::sort(double* Px, double* Py, int n)
 * ordonne lexico P selon un tri hyper lent [beurk]
 *****/
void clusters::sort(double* Px, double* Py, int n)
{
  int i,j;
  double t;
  if (n>1){
    for (i=0;i<n-1;i++){
      for (j=i+1;j<n;j++){
	if ( (Px[j]<Px[i]) || ( (Px[j]==Px[i]) && (Py[j]<Py[i]) ) ){
	  t=Px[j];
	  Px[j]=Px[i];
	  Px[i]=t;

	  t=Py[j];
	  Py[j]=Py[i];
	  Py[i]=t;
	}
      }
    }
  }
}

/*****
 * int clusters::chainHull_2D(double* Px, double* Py, int n, double* Hx, double* Hy)
 * Input:  P[] = an array of 2D points
 *                   presorted by increasing x- and y-coordinates
 *             n = the number of points in P[]
 *     Output: H[] = an array of the convex hull vertices (max is n)
 *     Return: the number of points in H[]
 *****/
int clusters::chainHull_2D(double* Px, double* Py, int n, double* Hx, double* Hy)
{
    // the output array H[] will be used as the stack
    int    bot=0, top=(-1);  // indices for bottom and top of the stack
    int    i;                // array scan index

    // Get the indices of points with min x-coord and min|max y-coord
    int minmin = 0, minmax;
    float xmin = Px[0];
    for (i=1; i<n; i++)
      if (Px[i] != xmin) break;
    minmax = i-1;
    if (minmax == n-1) {       // degenerate case: all x-coords == xmin
      Hx[++top]= Px[minmin];
      Hy[top]  = Py[minmin];
        if (Py[minmax] != Py[minmin]){ // a nontrivial segment
	  Hx[++top] = Px[minmax];
	  Hy[top] = Py[minmax];
	}
	Hx[++top] = Px[minmin];  
        Hy[top] = Py[minmin];           // add polygon endpoint
        return top+1;
    }

    // Get the indices of points with max x-coord and min|max y-coord
    int maxmin, maxmax = n-1;
    float xmax = Px[n-1];
    for (i=n-2; i>=0; i--)
        if (Px[i] != xmax) break;
    maxmin = i+1;

    // Compute the lower hull on the stack H
    Hx[++top] = Px[minmin];      // push minmin point onto stack
    Hy[top] = Py[minmin];
    i = minmax;
    while (++i <= maxmin)
    {
      // the lower line joins P[minmin] with P[maxmin]
      if (isLeft( Px[minmin], Py[minmin], Px[maxmin], Py[maxmin], Px[i], Py[i]) >= 0 && i < maxmin)
	continue;          // ignore P[i] above or on the lower line

      while (top > 0)        // there are at least 2 points on the stack
        {
	  // test if P[i] is left of the line at the stack top
	  if (isLeft( Hx[top-1], Hy[top-1], Hx[top], Hy[top], Px[i], Py[i]) > 0)
	    break;         // P[i] is a new hull vertex
	  else
	    top--;         // pop top point off stack
        }
      Hx[++top] = Px[i];
      Hy[top] = Py[i];       // push P[i] onto stack
    }

    // Next, compute the upper hull on the stack H above the bottom hull
    if (maxmax != maxmin){      // if distinct xmax points
      Hx[++top] = Px[maxmax];
      Hy[top] = Py[maxmax];  // push maxmax point onto stack
    }
    bot = top;                 // the bottom point of the upper hull stack
    i = maxmin;
    while (--i >= minmax)
    {
      // the upper line joins P[maxmax] with P[minmax]
      if (isLeft( Px[maxmax], Py[maxmax], Px[minmax], Py[minmax], Px[i], Py[i]) >= 0 && i > minmax)
	continue;          // ignore P[i] below or on the upper line

      while (top > bot)    // at least 2 points on the upper stack
        {
	  // test if P[i] is left of the line at the stack top
	  if (isLeft( Hx[top-1], Hy[top-1], Hx[top], Hy[top], Px[i], Py[i]) > 0)
	    break;         // P[i] is a new hull vertex
	  else
	    top--;         // pop top point off stack
        }
      Hx[++top] = Px[i];
      Hy[top] = Py[i];       // push P[i] onto stack
    }
    if (minmax != minmin){
      Hx[++top] = Px[minmin];
      Hy[top] = Py[minmin];  // push joining endpoint onto stack
    }
    return top+1;
}

/*****
 * char *clusters::itoa(int i)
 * passe l'entier i dans une chaine de char
 *****/
char *clusters::itoa(int i)
{
  char *pos = BUF + sizeof(BUF) - 1;
  unsigned int u;
  int negative = 0;
  if (i < 0){
    negative = 1;
    u = ((unsigned int)(-(1+i))) + 1;
  } else {
    u = i;
  }
  *pos = 0;
  do {
    *--pos = '0' + (u % 10);
    u /= 10;
  } while (u);
  if (negative) {
    *--pos = '-';
  }
  return pos;
}

/*****
 * void cluster::sauvegarde_param(char a, char* entete, int iter)
 * sauvegarde L_KMN et Variabilite
 * entete est la chaine placee en debut de fichier
 *****/
void clusters::sauvegarde_param(char a, char* entete, int iter)
{
  FILE* fichier;
  char ns[255];
  strcpy(ns,sfile);
  strcat(ns,"_");
  strcat(ns,smethod);
  strcat(ns,"_params.txt");

  if (iter==0) printf("%s (#dim=%d #pts=%d #gen=%d #K_th=%d #K_ex=%d #Init=%s) [",entete,dimension,NP,
		      pointSet->vgeneration(),pointSet->vKC_th(),KC,(INITIALISATION_FORGY==1) ? "Forgy" : "Rando");

  fichier=fopen(ns,"a");
  if (iter==0) fprintf(fichier,"----------------------------------\n");
  if (iter==0) fprintf(fichier,"%s :\n",entete);
  if (iter==0) fprintf(fichier,"#dim\t #pts\t #gen\t #bal\t #K_th\t #K_ex\t #Init\n");
  if (iter==0) fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %s\n\n",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC,(INITIALISATION_FORGY==1) ? "Forgy" : "Rando");
  if (iter==0) fprintf(fichier,"c\t Iter\t L_KMN\n");
  fprintf(fichier,"%c\t %d\t %g\n", a, iter, L_KMN);
  fclose(fichier);
}

/*****
 * void cluster::sauvegarde_costs_against(clusters* wcap)
 * sauvegarde L_KMN pour le cluster recevant et le cluster en parametre,
 * et calcule le gain relatif de L_KMN pour wcap par rapport au clusters
 * recevant.
 *****/
void clusters::sauvegarde_costs_against(clusters* wcap)
{
  FILE* fichier;
  char cp;
  char ns[255];
  strcpy(ns,sfile);
  strcat(ns,"_comp_loss.txt");

  double myL = Min_L_KMN;
  double hisL = wcap->vMin_L_KMN();
  double relgain = (hisL - myL) / ( hisL + myL );

  if (relgain > PERC_SIGNIFICATIF)
    cp = '-';
  else if (relgain < -PERC_SIGNIFICATIF)
    cp = '+';
  else
    cp = '0';

  fichier=fopen(ns,"a");
  fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %d\t =>\t ",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC,INITIALISATION_FORGY);
  fprintf(fichier,"%12.10g %12.10g\t %8.6g\n", myL, hisL, relgain);
  fclose(fichier);

  /*strcpy(ns,sfile);
    strcat(ns,"_comp_loss_WTL.txt");
  
    fichier=fopen(ns,"a");
    fprintf(fichier,"%c",cp);

    fclose(fichier);
  */
}

/*****
 * void cluster::sauvegarde_props_against(clusters* wcap)
 * sauvegarde Min_Prop pour le cluster recevant et le cluster en parametre,
 * et calcule le gain relatif de Min_Prop pour wcap par rapport au clusters
 * recevant.
 *****/
void clusters::sauvegarde_props_against(clusters* wcap)
{
  FILE* fichier;
  char ns[255];
  char cp;
  strcpy(ns,sfile);
  strcat(ns,"_comp_prop.txt");

  double myL = Min_Prop;
  double hisL = wcap->vMin_Prop();
  double relgain = (hisL - myL) / ( hisL + myL );

  if (relgain > PERC_SIGNIFICATIF)
    cp = '-';
  else if (relgain < -PERC_SIGNIFICATIF)
    cp = '+';
  else
    cp = '0';

  fichier=fopen(ns,"a");
  fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %d\t =>\t ",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC,INITIALISATION_FORGY);
  fprintf(fichier,"%8.6g %8.6g\t %8.6g\n", myL, hisL, relgain);
  fclose(fichier);

  /*strcpy(ns,sfile);
    strcat(ns,"_comp_prop_WTL.txt");
    
    fichier=fopen(ns,"a");
    fprintf(fichier,"%c",cp);
    fclose(fichier);
  */
}

/*****
 * void cluster::sauvegarde_average_props_against(clusters* wcap)
 * sauvegarde Log_Prop pour le cluster recevant et le cluster en parametre,
 * et calcule le gain relatif de Log_Prop pour wcap par rapport au clusters
 * recevant.
 * Log_Prop est la proportion de points pour lesquels le log de la fonction
 * de cout a la fin est inferieur a la moyenne de tous les precedents: c'est
 * donc aussi la proportion de points pour lesquels le C (dans EXP(-C)) est
 * superieur a la moyenne des precedents: selon la theorie, cette quantite pourrait
 * decroitre exponentiellement
 *****/
void clusters::sauvegarde_average_props_against(clusters* wcap)
{
  FILE* fichier;
  char ns[255];
  char cp;
  strcpy(ns,sfile);
  strcat(ns,"_comp_average_prop.txt");

  double myL = Log_Prop;
  double hisL = wcap->vLog_Prop();
  double relgain = (hisL - myL) / ( hisL + myL );

  if (relgain > PERC_SIGNIFICATIF)
    cp = '-';
  else if (relgain < -PERC_SIGNIFICATIF)
    cp = '+';
  else
    cp = '0';

  fichier=fopen(ns,"a");
  fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %d\t =>\t ",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC,INITIALISATION_FORGY);
  fprintf(fichier,"%8.6g %8.6g\t %8.6g\n", myL, hisL, relgain);
  fclose(fichier);

  /*strcpy(ns,sfile);
    strcat(ns,"_comp_average_prop_WTL.txt");
    
    fichier=fopen(ns,"a");
    fprintf(fichier,"%c",cp);
    fclose(fichier);
  */
}

/*****
 * void cluster::sauvegarde_missed_against(clusters* wcap)
 * sauvegarde Prop_Missed_Fin_Algo pour le cluster recevant et le cluster en parametre,
 * et calcule le gain relatif de Prop_Missed_Fin_Algo pour wcap par rapport au clusters
 * recevant.
 *****/
void clusters::sauvegarde_missed_against(clusters* wcap)
{
  FILE* fichier;
  char ns[255];
  char cp;
  strcpy(ns,sfile);
  strcat(ns,"_comp_missed.txt");

  double myL = Prop_Missed_Fin_Algo;
  double hisL = wcap->vProp_Missed_Fin_Algo();
  double relgain = (hisL - myL) / ( hisL + myL );

  fichier=fopen(ns,"a");
  fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %d\t =>\t ",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC,INITIALISATION_FORGY);
  fprintf(fichier,"%8.6g %8.6g\t", myL, hisL);
  if ( (hisL + myL) == 0){
      fprintf(fichier,"0\n");
      cp = '0';
  }
  else{
    if (relgain > PERC_SIGNIFICATIF)
      cp = '-';
    else if (relgain < -PERC_SIGNIFICATIF)
      cp = '+';
    else
      cp = '0';

    fprintf(fichier,"%g\n", relgain);
  }
  fclose(fichier);

  /*strcpy(ns,sfile);
    strcat(ns,"_comp_missed_WTL.txt");
    
    fichier=fopen(ns,"a");
    fprintf(fichier,"%c",cp);
    fclose(fichier);
  */
}


/*****
 * void cluster::sauvegarde_max_loss_against(clusters* wcap)
 * sauvegarde Max_Loss pour le cluster recevant et le cluster en parametre,
 * (calcule en fin d'iteration)
 * et calcule le gain relatif de Max_Loss pour wcap par rapport au clusters
 * recevant.
 *****/
void clusters::sauvegarde_max_loss_against(clusters* wcap)
{
  FILE* fichier;
  char ns[255];
  char cp;
  strcpy(ns,sfile);
  strcat(ns,"_comp_max_loss.txt");

  double myL = Max_Loss;
  double hisL = wcap->vMax_Loss();
  double relgain = (hisL - myL) / ( hisL + myL );

  fichier=fopen(ns,"a");
  fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %d\t =>\t ",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC,INITIALISATION_FORGY);
  fprintf(fichier,"%8.6g %8.6g\t", myL, hisL);
  if ( (hisL + myL) == 0){
      fprintf(fichier,"0\n");
      cp = '0';
  }
  else{
    if (relgain > PERC_SIGNIFICATIF)
      cp = '-';
    else if (relgain < -PERC_SIGNIFICATIF)
      cp = '+';
    else
      cp = '0';

    fprintf(fichier,"%g\n", relgain);
  }
  fclose(fichier);

  /*strcpy(ns,sfile);
    strcat(ns,"_comp_max_loss_WTL.txt");
    
    fichier=fopen(ns,"a");
    fprintf(fichier,"%c",cp);
    fclose(fichier);
  */
}



/*****
 * void cluster::sauvegarde_scatter(void)
 * sauvegarde L_KMN
 *****/
void clusters::sauvegarde_scatter(void)
{
  FILE* fichier;
  char ns[255];
  strcpy(ns,sfile);
  strcat(ns,"_");
  strcat(ns,smethod);
  strcat(ns,"_scatter.txt");
  fichier=fopen(ns,"a");
  fprintf(fichier,"%g\n", L_KMN);
  fclose(fichier);
}

/*****
 * void cluster::sauvegarde_hits(char* entete, int it)
 * sauvegarde le nombre de clusters theoriques touches
 *****/
void clusters::sauvegarde_hits(char* entete, int it)
{
  FILE* fichier;
  char ns[255];
  strcpy(ns,sfile);
  strcat(ns,"_");
  strcat(ns,smethod);
  strcat(ns,"_hits.txt");
  fichier=fopen(ns,"a");
  if (it==0) fprintf(fichier,"%s :\n",entete);
  if (it==0) fprintf(fichier,"#dim\t #pts\t #gen\t #bal\t #K_th\t #K_ex\t #Init\n");
  if (it==0) fprintf(fichier,"%d\t %d\t %d\t %d\t %d\t %d\t %s\n\n",dimension,NP,
		       pointSet->vgeneration(),pointSet->vbalance(),pointSet->vKC_th(),KC, (INITIALISATION_FORGY==1) ? "Forgy" : "Rando" );
  if (it==0) fprintf(fichier,"Iter\t %Clusters missed\n");
  fprintf(fichier,"%d\t %g\n", it, count_prop_missed_clusters());
  fclose(fichier);
}

/*****
 * void cluster::sauvegarde_im(int iter)
 * sauvegarde l'image courante
 *****/
void clusters::sauvegarde_im(int iter)
{
  FILE* fichier;
  char ns[255];
  strcpy(ns,sfile);
  strcat(ns,"_");
  strcat(ns,smethod);
  strcat(ns,"_image_");
  strcat(ns,itoa(iter));
  strcat(ns,".tif");
  
  /*if (iter==8){
    int* card=new int[KC];
    int i;
    for (i=0;i<KC;i++) card[i]=0;
    for (i=0;i<NP;i++) card[clusterEmpirique[i]]++;
    for (i=0;i<KC;i++) printf("(%d,%d) ",i,card[i]);
    printf("\n");
    }*/

  if (iter==8) AFFICHE=VRAI;

  imageC* im;
  im=new imageC(ns,t_im,t_im);
  claque_im(im);
  im->sauvegarde();

  delete im;
}

/*****
 * void cluster::sauvegarde_im_theorique(void)
 * sauvegarde le clustering theorique
 *****/
void clusters::sauvegarde_im_theorique(void)
{
  FILE* fichier;
  char ns[255];
  strcpy(ns,sfile);
  strcat(ns,"_");
  strcat(ns,smethod);
  strcat(ns,"_theorique.tif");
  
  imageC* im;
  im=new imageC(ns,t_im,t_im);
  pointSet->claque_im(im,VRAI);
  im->sauvegarde();

  delete im;
}

/*****
 * double clusters::count_prop_missed_clusters(void)
 * renvoie la proportion de clusters theoriques qui ne comptent aucun
 * centre experimental dans leur cellule de Voronoi
 *****/
double clusters::count_prop_missed_clusters(void)
{
  int i,j,k,nct,KCT;
  KCT = pointSet->vKC_th();
  int* hits=new int[KCT];
  double dist, distmin;
  double vr;

  vr=0;
  for (i=0;i<KCT;i++) hits[i]=0;
  
  for (i=0;i<KC;i++){//num cluster estime
    if (getClusterValide(i)==VRAI){
      for (j=0;j<KCT;j++){//nc theorique
	dist=0;
	for (k=0;k<dimension;k++){
	  dist+=( ((double) pointSet->getCoordCentre(j,k) - getCoord(centersCoordFromClusters,i,k)) )
	    *( ((double) pointSet->getCoordCentre(j,k) - getCoord(centersCoordFromClusters,i,k)) );
	}
	if ( (j==0) || (dist<distmin) ){
	  nct=j;
	  distmin=dist;
	}
      }
      hits[nct]++;
    }
  }

  for (i=0;i<KCT;i++) vr+=(hits[i]>0) ? 1 : 0;

  vr = ( ( (double) KCT ) - vr )/( (double) KCT );

  return vr;
}


/*****
 * void clusters::put_dist(int v)
 * remplace la distance
 *****/
void clusters::put_dist(int v)
{
  DISTANCE=v;
  if ( (DISTANCE==0) || (DISTANCE==2) || (DISTANCE==3) ) NIVELLAGE=1;
  else if (DISTANCE==1) NIVELLAGE=CONST_MOY_LIMITE;
  else exit(0);
}


/*****
 * double clusters::distance_pt_centre(int pt, int centre, double *tabCentres)
 * renvoie la distance euclidienne entre un pt de pointSet et un centre
 *****/
double clusters::distance_pt_centre(int pt, int centre, double *tabCentres)
{
  double v=0;
  int i;

  for (i=0;i<dimension;i++) 
    v+=( (double) pointSet->getCoordPoints(pt,i)-getCoord(tabCentres,centre,i))
      *( (double) pointSet->getCoordPoints(pt,i)-getCoord(tabCentres,centre,i));
  
  //v=sqrt(v);
  
  return v;
}

/*****
 * double clusters::KMN_Loss_KMeans(void)
 * calcule et renvoie le cout KMN des KMeans
 *****/
double clusters::KMN_Loss_KMeans(void)
{
  double cout =0;
  double cco;
  int i;
  for (i=0;i<NP;i++){
    cco=Pointwise_KMN_Loss_KMeans(i);
    cout+=cco;
  } 
  //cout = sqrt(cout);
  cout /= (double) NP;
  return cout;
}

/*****
 * double clusters::KMN_Loss_KMeans_Without_GetMembership(void)
 * calcule et renvoie le cout KMN des KMeans
 * Mais n'utilise pas la fonction getMembership(): on recalcule completement
 * l'appartenance du point
 *****/
double clusters::KMN_Loss_KMeans_Without_GetMembership(void)
{
  double cout =0;
  double cco;
  int i;
  for (i=0;i<NP;i++){
    cco=Pointwise_KMN_Loss_KMeans_Without_GetMembership(i);
    cout+=cco;
  } 
  //cout = sqrt(cout);
  cout /= (double) NP;
  return cout;
}

/*****
 * double clusters::Pointwise_KMN_Loss_KMeans(int p)
 * calcule et renvoie le cout KMN des KMeans pour un point
 *****/
double clusters::Pointwise_KMN_Loss_KMeans(int p)
{
  //On implemente cette fois vraiment le KMN Loss, comme voulu dans Elkan et Cie
  
  int i;
  double v;
  int nc=-1;
  
  for (i=0;i<KC;i++){
    if (getMembership(i,p)==1){
      nc = i;
    }
  }
  v = distance_pt_centre(p,nc,centersCoordFromClusters);

  assert(nc != -1);
  return v;
}

/*****
 * double clusters::Pointwise_KMN_Loss_KMeans_Without_GetMembership(int p)
 * calcule et renvoie le cout KMN des KMeans pour un point
 * Mais n'utilise pas la fonction getMembership(): on recalcule completement
 * l'appartenance du point
 *****/
double clusters::Pointwise_KMN_Loss_KMeans_Without_GetMembership(int p)
{
  double dist, distmin;
  BOOL trouve;
  int k, nc, nval;

  trouve=FAUX;
  nc=-1;
  nval=0;
  for (k=0;k<KC;k++){
    //printf("%d ",k);
    if (getClusterValide(k)==VRAI){
      nval++;
      dist=distance_pt_centre(p,k,centersCoordFromClusters);
      //printf("%g ",dist);
      if ( (trouve==FAUX) || (dist<distmin) ) {
	nc=k;
	distmin=dist;
	trouve=VRAI;
      }
    }
  }

  if (nval == 0) printf("\n\nATTENTION :: Aucun Cluster !!\n");
  
  assert(nc!=-1);

  return distmin;
}

/*****
 * double clusters::Pointwise_HM_Loss(int p)
 * calcule et renvoie le cout Harmonic Means pour un point
 *****/
double clusters::Pointwise_HM_Loss(int p)
{
  int k;
  double v=(double) KC;
  double den=0;
  double distc;
  int nv=0;

  for (k=0;k<KC;k++){
    if (getClusterValide(k)==VRAI){
      nv++;
      distc = distance_pt_centre(p,k,centersCoordFromClusters);
      den += (distc < EPSILON_KHM) ? (1/EPSILON_KHM) : (1/distc) ;
    }
  }
  
  v = v / den;

  //printf("%g\n",v);

  if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
  assert(nv>0);

  return v;
}


/*****
 * double clusters::Harmonic_Means_Loss(void)
 * calcule et renvoie le cout Harmonique moyen
 *****/
double clusters::Harmonic_Means_Loss(void)
{
  double cout =0;
  double cco;
  int i;
  for (i=0;i<NP;i++){
    cco=Pointwise_HM_Loss(i);
    cout+=cco;
  } 
  //cout = sqrt(cout);
  cout /= (double) NP;
  return cout;
}


/*****
 * void clusters::Compute_L_HM(void)
 * calcule L_HM
 *****/
void clusters::Compute_L_HM(void)
{
  L_KMN=Harmonic_Means_Loss();
}

/*****
 * void clusters::Compute_L_KMN(void)
 * calcule L_KMN
 *****/
void clusters::Compute_L_KMN(void)
{
  L_KMN=KMN_Loss_KMeans();
}

/*****
 * void clusters::Compute_L_KMN_Without_GetMembership(void)
 * calcule L_KMN
 * Mais n'utilise pas la fonction getMembership(): on recalcule completement
 * l'appartenance du point
 *****/
void clusters::Compute_L_KMN_Without_GetMembership(void)
{
  L_KMN=KMN_Loss_KMeans_Without_GetMembership();
}

/*****
 * void clusters::init_clusters_centers_Forgy(void)
 * init aleatoire des centres selon E. Forgy
 *****/
void clusters::init_clusters_centers_Forgy(void)
{
  int i,j,k;

  for (i=0;i<KC;i++){
    j=abs(rand()%NP);
    //printf("%d %d\n",i,j);
    //pointSet->affiche();
    for (k=0;k<dimension;k++){
      //printf("k=%d %g\n",k,(double) pointSet->getCoordPoints(j,k));
      putCoord(centersCoordFromClusters,i,k,(double) pointSet->getCoordPoints(j,k));
    }
  }
}

/*****
 * void clusters::init_clusters_centers_Random(void)
 * init aleatoire des centres selon Random Partition
 *****/
void clusters::init_clusters_centers_Random(void)
{
  int i,j,k;
  int* member=new (int) [NP];
  double* cardP=new (double) [KC];
  double coord;

  printf("Generating clusters... ");
  for (i=0;i<KC;i++) cardP[i]=0;
  for (i=0;i<NP;i++){
    member[i]=abs(rand()%KC);
    cardP[member[i]]++;
  }

  for (i=0;i<KC;i++){
    if (cardP[i]==0){
      clusterValide[i] = FAUX;
      printf("\nWarning: in the random Generation, no point has been affected to cluster #%d\n", i);
    }
    else{
      for (k=0;k<dimension;k++){
	coord=0;
	for (j=0;j<NP;j++){
	  if (member[j]==i){
	    coord+=(double) pointSet->getCoordPoints(j,k);
	  }
	}
	coord/=cardP[i];
	putCoord(centersCoordFromClusters,i,k,coord);
      }
    }
  }
  printf("...end of cluster generation\n");

  delete [] cardP;
  delete [] member; 
}

/*****
 * void clusters::affecte_points_clusters_Hard_Membership(void)
 * affecte les points aux clusters, sachant que les centres sont choisis
 * ATTENTION: la methode est celle du Hard Membership
 *****/
void clusters::affecte_points_clusters_Hard_Membership(void)
{
  int j,k,nc,nval=0;
  double dist, distmin;
  BOOL trouve=FAUX;

  for (j=0;j<NP;j++)
    for (k=0;k<KC;k++)
      putMembership(k,j,-1);
  //clusterEmpirique[j]=-1;

  for (j=0;j<NP;j++){
    trouve=FAUX;
    nc=-1;
    nval=0;
    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	nval++;
	dist=distance_pt_centre(j,k,centersCoordFromClusters);
	if ( (trouve==FAUX) || (dist<distmin) ) {
	  nc=k;
	  distmin=dist;
	  trouve=VRAI;
	}
      }
    }
    for (k=0;k<KC;k++){
      if (k==nc)
	putMembership(k,j,1);
      else
	putMembership(k,j,0);
    }

    if (nval == 0) printf("\n\nATTENTION :: Pas assez de points ou de clusters pour l'initialisation choisie : arret.\n");

    assert(nc!=-1);
  }
}

/*****
 * void clusters::membership_KHM(void)
 * affecte les points aux clusters, sachant que les centres sont choisis
 * ATTENTION: la methode est celle du Harmonic Means Clustering
 *****/
void clusters::membership_KHM(void)
{
  int j,k,nv;
  double distc, contribc, contriball;

  for (j=0;j<NP;j++){
    contriball=0;
    nv = 0;
    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	nv++;
	distc = distance_pt_centre(j,k,centersCoordFromClusters);
	contribc = 1/ppd(distc);
	contriball += contribc;
      }
    }
    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	distc = distance_pt_centre(j,k,centersCoordFromClusters);
	contribc = 1/ppd(distc);
	putMembership(k,j,contribc/contriball);
	//printf("[%d %g]",k,contribc/contriball);
      }
      else
	putMembership(k,j,0);
    }
    //printf("\n");

    if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
    assert(nv>0);
  }
}

/*****
 * void clusters::membership_WKHM(void)
 * affecte les points aux clusters, sachant que les centres sont choisis
 * ATTENTION: la methode est celle du Weighted Harmonic Means Clustering,
 * AU SENS DE HAMMERLY ET ELKAN
 *****/
void clusters::membership_WKHM(void)
{
  //affecte_points_clusters_KHM();
  double* sumden=new double [NP];
  double distc, cont, gi;
  int j,k;
  
  for (j=0;j<NP;j++){
    sumden[j]=0;
    for (k=0;k<KC;k++){
      assert(getClusterValide(k)==VRAI);
      distc = distance_pt_centre(j,k,centersCoordFromClusters);
      sumden[j] += (1/p(distc));
    }
    sumden[j] = (sumden[j] * sumden[j]);
  }

  for (k=0;k<KC;k++){
    for (j=0;j<NP;j++){
      distc = distance_pt_centre(j,k,centersCoordFromClusters);
      gi = (1/ppd(distc)) / sumden[j];
      putMembership(k,j,gi);
    }
  }
}

/*****
 * void clusters::calcule_centres_clusters(void)
 * recalcule les centres de clusters, sur la base du vecteur de poids
 * pour les exemples, selon la methode des moindre carres
 *****/
void clusters::calcule_centres_clusters(void)
{
  int i,j,k;
  double sav, sap;
  double* denomK=new (double) [KC];
  double* calcvalid=new (double) [KC];

  for (j=0;j<KC;j++){
    denomK[j]=0;
    calcvalid[j]=0;
  }
  for (i=0;i<KC;i++)
    for (j=0;j<dimension;j++) 
      putCoord(centersCoordFromClusters,i,j,0);
    
  for (i=0;i<NP;i++){
    for (k=0;k<KC;k++){
      //printf("Calcule_Centre_Clusters: (%d,%d) : %d,(W=%g,M=%g)\n",i,k,getClusterValide(k),weights[i],getMembership(k,i));
      if (getClusterValide(k)==VRAI){
	calcvalid[k]+=getMembership(k,i);
	denomK[k]+=weights[i]*getMembership(k,i);
	for (j=0;j<dimension;j++){
	  sav = getCoord(centersCoordFromClusters,k,j);
	  sap = sav + ( (double) pointSet->getCoordPoints(i,j)*(weights[i]*getMembership(k,i)));
	  putCoord(centersCoordFromClusters,k,j,sap);
	}
      }
    }
  }
  
  for (i=0;i<KC;i++){
    if (getClusterValide(i)==VRAI){
      if (calcvalid[i]>0){
	for (j=0;j<dimension;j++){
	  sap = (getCoord(centersCoordFromClusters,i,j)/denomK[i]);
	  putCoord(centersCoordFromClusters,i,j,sap);
	}
      }
      else{
	putClusterValide(i,FAUX);
	printf("[Removing cluster %d]",i);
      }
    }
  }

  delete [] calcvalid;
  delete [] denomK;
}

/*****
 * void clusters::Generic_K_Means(int nstep, int typec)
 * algorithme des (weighted) k-means, 2005
 * typec = 0 -> la valeur de c est fixee par la nouvelle formule
 * typec = 1 -> la valeur de c est calculee par dichotomies
 * typec = 2 -> c'est l'algo des k-means traditionnel qui est execute
 *****/
void clusters::Generic_K_Means(int nstep, int typec){
  if (typec == 0){
    strcpy(smethod,"WKM_C_Constant");
  }else if (typec == 1){
    strcpy(smethod,"WKM_C_Dicho");
  }else if (typec == 2){
    strcpy(smethod,"KM");
  }
  //printf("%s",smethod); 

  sauvegarde_im_theorique();

  char ns[255];
  if (typec == 0)
    strcpy(ns,"Weighted_K_Means_C_Exact");
  else if (typec == 1)
    strcpy(ns,"Weighted_K_Means_C_Dicho");
  else if (typec == 2)
    strcpy(ns,"K_Means");

  int i, step;
  double dmaxt, dplust, dmoinst, lc, abl, ct, Z;
  double prop;

  double* Log_Loss_Init=new (double) [NP];
  double* Average_Log_Loss=new (double) [NP];
  double fraction;

  for (i=0;i<NP;i++){
    weights[i]=(double)1/(double) NP;
    Average_Log_Loss[i]=0;
  }

  //sauvegarde_im(-1);
  for (step=0;step<nstep;step++){

    if (typec<=2)
      affecte_points_clusters_Hard_Membership();

    if (step==0)
      for (i=0;i<NP;i++){
	Log_Loss_Prec[i] = Pointwise_KMN_Loss_KMeans(i);
	Log_Loss_Init[i] = Log_Loss_Prec[i];
      }
    else{
      for (i=0;i<NP;i++){
	Log_Loss_Cour[i] = Pointwise_KMN_Loss_KMeans(i);
      }
    }

    if (step==0){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Init[i];
    }else if (step<nstep-1){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Cour[i];
    }

    if ( (step>0) && (typec < 2) )
      ct=weights_abstract(Log_Loss_Prec, Log_Loss_Cour, typec, step);

    //calculs divers
    Compute_L_KMN();
    if ( (step==0) || (L_KMN < Min_L_KMN) )
      Min_L_KMN = L_KMN;

    if (step==nstep-1){ //ATTENTION: CHANGEMENT
      fraction=0;
      for (i=0;i<NP;i++){
	if (Log_Loss_Init[i] <= Log_Loss_Cour[i])
	  fraction++;
      }
      fraction/=(double) NP;
      //if (fraction < Min_Prop)
      Min_Prop = fraction;
    }

    //sauvegardes diverses
    
    if (step>=0){
      if (typec < 2)
	if (ct<0) sauvegarde_param('-',ns,step);
	else if (ct==0) sauvegarde_param('z',ns,step);
	else sauvegarde_param('+',ns,step);
      else
	sauvegarde_param('.',ns,step);
    }

    if (step>0)
      for (i=0;i<NP;i++) Log_Loss_Prec[i] = Log_Loss_Cour[i];

    if (step==nstep-1)
      Prop_Missed_Fin_Algo = count_prop_missed_clusters();

    //recalcul des centres
    calcule_centres_clusters();

    if (step%STEP_FOR_SAVING_IMG==0)
      sauvegarde_im(step);
    printf(".");
    fflush(stdout);
  }

  for (i=0;i<NP;i++)
    if ( (i==0) || (Log_Loss_Cour[i]>Max_Loss) )
      Max_Loss = Log_Loss_Cour[i];

  for (i=0;i<NP;i++){
    Average_Log_Loss[i]/=(double) (nstep-1);
  }

  fraction=0;
  for (i=0;i<NP;i++){
    if (Log_Loss_Cour[i] >= Average_Log_Loss[i])
      fraction++;
  }
  fraction/=(double) NP;
  Log_Prop = fraction;

  printf("]\n");
  sauvegarde_scatter();

  delete [] Log_Loss_Init;
  delete [] Average_Log_Loss;
}

/*****
 * void clusters::divide_loss(double* tab)
 * divise le cout par une constante, NORMALISATION_CONST
 *****/
void clusters::divide_loss(double* tab){
  int i;
  for (i=0;i<NP;i++){
    tab[i] = tab[i] / ( (double) NORMALISATION_CONST );
  }
}

/*****
 * void clusters::Generic_Harmonic_Means(int nstep, int typec)
 * algorithme des (weighted) HARMONIC k-means, 2005
 * typec = 0 -> la valeur de c est fixee par la nouvelle formule
 * typec = 1 -> la valeur de c est calculee par dichotomies
 * typec = 2 -> c'est l'algo des Harmonic-means traditionnel qui est execute
 *****/
void clusters::Generic_Harmonic_Means(int nstep, int typec){
  if (typec == 0){
    strcpy(smethod,"WHM_C_Constant");
  }else if (typec == 1){
    strcpy(smethod,"WHM_C_Dicho");
  }else if (typec == 2){
    strcpy(smethod,"HM");
  }

  sauvegarde_im_theorique();

  char ns[255];
  if (typec == 0)
    strcpy(ns,"Weighted_Harmonic_Means_C_Exact");
  else if (typec == 1)
    strcpy(ns,"Weighted_Harmonic_Means_C_Dicho");
  else if (typec == 2)
    strcpy(ns,"Harmonic_Means");

  int i, step;
  double dmaxt, dplust, dmoinst, lc, abl, ct, Z;
  double prop;

  double* Log_Loss_Init=new (double) [NP];
  double* Average_Log_Loss=new (double) [NP];
  double fraction;

  for (i=0;i<NP;i++) weights[i]=(double)1/(double) NP;

  for (step=0;step<nstep;step++){

    if (typec<2)
      membership_WKHM();
    else
      membership_KHM();

    if (step==0){
      for (i=0;i<NP;i++){
	Log_Loss_Prec[i] = Pointwise_HM_Loss(i);
	Log_Loss_Init[i] = Log_Loss_Prec[i];
      }
      //divide_loss(Log_Loss_Prec);
      //divide_loss(Log_Loss_Init);
    }
    else{
      for (i=0;i<NP;i++){
	Log_Loss_Cour[i] = Pointwise_HM_Loss(i);	
      }
      //divide_loss(Log_Loss_Cour);
      for (i=0;i<NP;i++){
	//printf("[%d] Prec = %g, Cour = %g /// %d \n",i,Log_Loss_Prec[i],Log_Loss_Cour[i],(Log_Loss_Prec[i]>Log_Loss_Cour[i]) ? 1 : 0);
      }
    }

    if (step==0){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Init[i];
    }else if (step<nstep-1){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Cour[i];
    }

    if ( (step>0) && ( typec < 2) )
      ct=weights_abstract(Log_Loss_Prec, Log_Loss_Cour, typec, step);

    if ( (step>0) && (typec == 2) )
      weights_KHM();

    //calculs divers
    Compute_L_KMN_Without_GetMembership();//Compute_L_HM();
    if ( (step==0) || (L_KMN < Min_L_KMN) )
      Min_L_KMN = L_KMN;

    if (step==nstep-1){ //ATTENTION: CHANGEMENT !!
      fraction=0;
      for (i=0;i<NP;i++){
	if (Log_Loss_Init[i] <= Log_Loss_Cour[i])
	  fraction++;
      }
      fraction/=(double) NP;
      //if (fraction < Min_Prop)
      Min_Prop = fraction;
    }

    //sauvegardes diverses
    
    if (step>=0){
      if (typec < 2)
	if (ct<0) sauvegarde_param('-',ns,step);
	else if (ct==0) sauvegarde_param('z',ns,step);
	else sauvegarde_param('+',ns,step);
      else
	sauvegarde_param('.',ns,step);
    }

    if (step>0)
      for (i=0;i<NP;i++) Log_Loss_Prec[i] = Log_Loss_Cour[i];

    if (step==nstep-1)
      Prop_Missed_Fin_Algo = count_prop_missed_clusters();

    //recalcul des centres
    calcule_centres_clusters();

    if (step%STEP_FOR_SAVING_IMG==0)
      sauvegarde_im(step);
    printf(".");
    fflush(stdout);
  }

  for (i=0;i<NP;i++)
    if ( (i==0) || (Log_Loss_Cour[i]>Max_Loss) )
      Max_Loss = Log_Loss_Cour[i];

  for (i=0;i<NP;i++){
    Average_Log_Loss[i]/=(double) (nstep-1);
  }

  fraction=0;
  for (i=0;i<NP;i++){
    if (Log_Loss_Cour[i] >= Average_Log_Loss[i])
      fraction++;
  }
  fraction/=(double) NP;
  Log_Prop = fraction;

  printf("]\n");
  sauvegarde_scatter();

  delete [] Log_Loss_Init;
  delete [] Average_Log_Loss;
}

/*****
 * double clusters::weights_abstract(double* Prec, double* Cour, int typec, int step)
 * calcule les poids des exemples dans l'abstract weighting scheme
 *****/
double clusters::weights_abstract(double* Prec, double* Cour, int typec, int step)
{
  double ct,Z,lc;
  int i;
  if (typec == 0)
    ct = c_fixe(Prec, Cour);
  else if (typec == 1){
    ct = c_dicho(Prec, Cour);
  }

  if (ct!=0)
    {
      Z=0;
      for (i=0;i<NP;i++){
	lc = Cour[i] - Prec[i];
	weights[i]*=exp(-ct*lc);
	Z+=weights[i];
      }
      for (i=0;i<NP;i++){
	weights[i]/=Z;
	if ((i==0) || (weights[i]>maxWeight)) maxWeight=weights[i];
	if ((i==0) || (weights[i]<minWeight)) minWeight=weights[i];
      }
    }

  return ct;
}

/*****
 * void clusters::weights_KHM(void)
 * calcule les poids des points
 * ATTENTION: la methode est celle du Harmonic Means Clustering
 *****/
void clusters::weights_KHM(void)
{
  int j,k,nv;
  double distc, num, den;

  for (j=0;j<NP;j++){
    num = 0;
    den = 0;
    nv = 0;
    for (k=0;k<KC;k++){
      assert(getClusterValide(k)==VRAI);
      nv++;
      distc = distance_pt_centre(j,k,centersCoordFromClusters);
      num += 1/ppd(distc);
      den += 1/p(distc);
    }
    weights[j]=num / (den*den);

    if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
    assert(nv>0);
  }
}

/*****
 * double clusters::Z_prime(double* Prec, double* Suiv, double c)
 * derivee de Z, calculee avec c
 *****/
double clusters::Z_prime(double* Prec, double* Suiv, double c)
{
  int i;
  double Zp = 0;

  for (i=0;i<NP;i++){
    Zp+=(-weights[i]*(Suiv[i]-Prec[i])*exp(-c*(Suiv[i]-Prec[i])));
  }

  return Zp;
}

/*****
 * double clusters::c_dicho(double* Prec, double* Suiv)
 * approximation de c par dichotomies
 *****/
double clusters::c_dicho(double* Prec, double* Cour)
{
  double dmaxt, dmint, dplust, dmoinst, abl, lc, cmin, cmax, Zpcmin, Zpcmax, Zpcour, ct, bigabs, rel;
  int i;
  BOOL stop = FAUX;
  BOOL pos = FAUX;
  BOOL arret_autorise = FAUX;

  //maj des poids
  dmaxt=0;
  dmint=0;
  dplust=0;
  dmoinst=0;
  for (i=0;i<NP;i++){
    assert(weights[i]>=0);

    lc = Cour[i] - Prec[i];
    abl = (lc>0) ? lc : -lc;
    if (abl > dmaxt)
      dmaxt = abl;
    if ( (dmint==0) || ( (abl > 0) && (abl < dmint) ) )
      dmint = abl;

    if (lc<0)
      dmoinst += weights[i]*lc;
    if (lc>0)
      dplust += weights[i]*lc;
  }
  if (dmoinst != 0)
    dmoinst = -dmoinst;

  assert(dmint > 0);

  //printf("%g, %g\n",dplust,dmoinst);

  if ( (dmoinst>0) && (dplust>0) ){

    if (dmoinst > dplust){
      //ct est negatif
      cmin = -(1/(2*dmint)) * log (dmoinst / dplust);
      cmax = -(1/(2*dmaxt)) * log (dmoinst / dplust);
      bigabs = -cmin;
    }
    else if (dmoinst < dplust){
      //ct est positif
      pos = VRAI;
      cmin = (1/(2*dmaxt)) * log (dplust / dmoinst);
      cmax = (1/(2*dmint)) * log (dplust / dmoinst);
      bigabs = cmax;
    }
    else{
      ct = 0;
      stop = VRAI;
    }
    
    //printf("cmin = %g, cmax = %g\n",cmin,cmax);
    
    double inc = 0.0001;
    double Zc;
    if ( (pos == FAUX) && (cmin < -0.1) ) {
      cmin = cmax;
      do{
	Zc = Z_prime(Prec, Cour, cmin);
	if (Zc > 0){
	  cmin -= inc;
	}
      }while(Zc>0);
      //printf("cmin=%g, cmax=%g, Zpcmin=%g\n",cmin,cmax,Zc);
    }
    
    if ( (pos == VRAI) && (cmax > 0.1) ) {
      cmax = cmin;
      do{
	Zc = Z_prime(Prec, Cour, cmax);
	if (Zc < 0){
	  cmax += inc;
	}
      }while(Zc<0);
      //printf("cmin=%g, cmax=%g, Zpcmax=%g\n",cmin,cmax,Zc);
    }
    
    while(stop == FAUX){
      ct = (cmax + cmin)/2;
      Zpcour = Z_prime(Prec, Cour, ct);
      rel = (cmax - cmin)/bigabs;
      if ( rel < LIMIT_PERCENTAGE ){
	if ( (pos == FAUX) && (Zpcour > 0) )
	  stop = VRAI;
	else if ( (pos == VRAI) && (Zpcour < 0) )
	  stop = VRAI;
	else if (Zpcour == 0)
	  stop = VRAI;
      }
      
      if ( stop == FAUX ){
	if (Zpcour > 0){
	  cmax = ct;
	  Zpcmax = Zpcour;
	}
	else{
	  cmin = ct;
	  Zpcmin = Zpcour;
	}
	
	if (pos == FAUX)
	  bigabs = -cmin;
	else
	  bigabs = cmax;
      }
    }
  }
  else if (dmoinst>0)
    ct = - C_MAX;
  else if (dplust>0)
    ct = C_MAX;
  else
    ct = 0;

  //printf("ct = %g, Zpct = %g\n",ct,Zpcour);

  return ct;
}

/*****
 * double clusters::c_fixe(double* Prec, double* Suiv)
 * renvoie la valeur de c sans recherche dicho, avec le nouveau resultat
 * Utilise la formule (18) du draft
 *****/
double clusters::c_fixe(double* Prec, double* Cour)
{
  int i;
  double num, den, dmaxt, ct, lc, abl;
  num = 0;
  den = 0;
  for (i=0;i<NP;i++){
    lc = Log_Loss_Cour[i] - Log_Loss_Prec[i];
    abl = (lc>0) ? lc : -lc;
    if ( (i==0) || (abl > dmaxt) )
      dmaxt = abl;
    num += weights[i]*lc;
    den -= weights[i]*lc;
  }
  num += dmaxt;
  den += dmaxt;

  ct = (1/(2*dmaxt)) * log ( (num + EPSILON_LOG_KHM) / (den + EPSILON_LOG_KHM) );
  if (dmaxt == 0)
    ct = 0; //pas de modif des poids, comme il n'y a pas de variation de fonction de cout

  //printf("ct = %g ; dmaxt = %g ; num = %g ; den = %g \n",ct,dmaxt,num,den);

  return ct;
}

/*****
 * void clusters::EM_init_mixingProp(void);
 * initialise mixingProp uniforme
 *****/
void clusters::EM_init_mixingProp(void){
  int i;
  for (i=0;i<KC;i++){
    mixingProp[i] = 1/ (double) KC;
  }
}

/*****
 * void clusters::EM_init_membership(void);
 * initialise membership uniforme
 *****/
void clusters::EM_init_membership(void){
  int i,j;
  for (i=0;i<NP;i++){
    for (j=0;j<KC;j++){
      putMembership(j,i,1/ (double) KC);
    }
  }
}

/*****
 * void clusters::EM_redispatche_membership(int clust)
 * on redispatche les memberships, sachant que clust
 * est supprime
 *****/
void clusters::EM_redispatche_membership(int clust){
  int i,k;
  double mem, bis, card,total;
  for (i=0;i<NP;i++){
    mem = getMembership(clust,i);

    putMembership(clust,i,0);

    total=0;
    for (k=0;k<KC;k++){
      if ( (getClusterValide(k)==VRAI) && (k!=clust) ){
	bis = getMembership(k,i);
	total+=bis;
      }
    }

    if ( (mem<1.0-TOUPITI) && (total>0) ) {

      assert(total>0);

      for (k=0;k<KC;k++){
	if ( (getClusterValide(k)==VRAI) && (k!=clust) ){
	  bis = getMembership(k,i);
	  putMembership(k,i,bis/total);
	}
      }
    }
    else{
      card=0;
      for (k=0;k<KC;k++){
	if ( (getClusterValide(k)==VRAI) && (k!=clust) ){
	  card++;
	}
      }
      assert(card>0);
      for (k=0;k<KC;k++){
	if ( (getClusterValide(k)==VRAI) && (k!=clust) ){
	  putMembership(k,i,1/card);
	}
      }
    }
  }
}

/*****
 * void clusters::EM_checkup(BOOL diagonal)
 * si un cluster est sous-represente, on le supprime
 *****/
void clusters::EM_checkup(BOOL diagonal){
  int i;
  for (i=0;i<KC;i++){
    if (getClusterValide(i)==VRAI){
      if (varcovar[i]->noZeroInDiagonal()==0){
	printf("[Removing cluster %d]",i);
	EM_redispatche_membership(i);
	putClusterValide(i,FAUX);
      }
    }
  }
}

/*****
 * void clusters::EM_approximate_checkup(BOOL diagonal)
 * si un cluster est sous-represente, on le supprime
 *****/
void clusters::EM_approximate_checkup(BOOL diagonal){
  int i;
  for (i=0;i<KC;i++){
    if (getClusterValide(i)==VRAI){
      if (varcovar[i]->approxZeroInDiagonal()==0){
	printf("[Removing cluster %d]",i);
	EM_redispatche_membership(i);
	putClusterValide(i,FAUX);
      }
    }
  }
}

/*****
 * void clusters::EM_crunch_Sigmas(BOOL diagonal)
 * calcule Sigmas, recopie dans les inverses, et calcule les inverses
 *****/
void clusters::EM_crunch_Sigmas(BOOL diagonal){
  EM_compute_Sigmas(diagonal);
  EM_checkup(diagonal);
  EM_Sigmas_To_Inverse();
  EM_compute_Inverse_Sigmas(diagonal);

  int i,test;
  int globtest = 0;
  /*for (i=0;i<KC;i++)
    varcovar[i]->affiche();

  for (i=0;i<KC;i++)
  inverseVarcovar[i]->affiche();*/

  /*for (i=0;i<KC;i++){
    printf("Cluster:%d\n",i);
    varcovar[i]->affiche();
    inverseVarcovar[i]->affiche();
    }*/
  for (i=0;i<KC;i++){
    if (getClusterValide(i)==VRAI){
      test=varcovar[i]->check_inverse(inverseVarcovar[i]);
      if (test==0)
	globtest = 1;
    }
  }

  if (globtest == 1){
    EM_approximate_checkup(diagonal);
    EM_Sigmas_To_Inverse();
    EM_compute_Inverse_Sigmas(diagonal);
    
    for (i=0;i<KC;i++){
      if (getClusterValide(i)==VRAI){
	test=varcovar[i]->check_inverse(inverseVarcovar[i]);

	assert(test == 1);
      }
    }
  }
}

/*****
 * void clusters::EM_compute_Sigmas(BOOL diagonal)
 * calcule les Sigmas
 * ATTENTION: les centres sont supposes calcules
 *****/
void clusters::EM_compute_Sigmas(BOOL diagonal){
  int i,j,row,col;
  double valcour, norm, crow, ccol, murow, mucol, prod, sum;
  for (j=0;j<KC;j++){
    if (getClusterValide(j)==VRAI){
      for (row=0;row<dimension;row++){
	for (col=0;col<dimension;col++){
	  norm = 0;
	  sum = 0;
	  for (i=0;i<NP;i++){
	    assert(weights[i]>=0);
	    assert(getMembership(j,i)>=0);
	    
	    prod = weights[i]*getMembership(j,i); //weights[i];//getMembership(j,i);//weights[i]*getMembership(j,i);
	    
	    //if (j==1) printf ("Member(%d)=%g\n",i,prod);
	    
	    crow = pointSet->getCoordPoints(i,row);
	    ccol = pointSet->getCoordPoints(i,col);
	    
	    murow = getCoordCenter(j,row);
	    mucol = getCoordCenter(j,col);
	    
	    valcour = prod*(crow - murow)*(ccol - mucol);
	    
	    if ( (row != col) && (diagonal==VRAI) ){
	      valcour = 0;
	    }
	    sum += valcour;
	    norm += prod;
	  }
	  //on met la coord
	  assert(norm>0);
	putSigma(row,col,j,(sum/norm));
	}
      }
    }
  }
}

/*****
 * void clusters::EM_Sigmas_To_Inverse(void)
 * recopie les Sigmas dans les Inverses (avant de calculer les inverses)
 *****/
void clusters::EM_Sigmas_To_Inverse(void){
  int j,row,col;
  double val;
  for (j=0;j<KC;j++){
    if (getClusterValide(j)==VRAI){
      for (row=0; row<dimension; row++){
	for (col=0;col<dimension; col++){
	  val = getSigma(row,col,j);
	  putInvSigma(row,col,j,val);
	}
      }
    }
  }
}

/*****
 * void clusters::EM_compute_Inverse_Sigmas(BOOL diagonal)
 * calcule les inverse des Sigmas, ainsi que det_exp et det_man
 * ATTENTION: les Sigmas sont supposes calcules !
 *****/
void clusters::EM_compute_Inverse_Sigmas(BOOL diagonal){
  int j;
  int res;
  double m;
  int e;
  EM_Sigmas_To_Inverse();
  for (j=0;j<KC;j++){
    if (getClusterValide(j)==VRAI){
      if (diagonal == FAUX)
	res = inverseVarcovar[j]->inverse(&m,&e);
      else
	res = inverseVarcovar[j]->inverse_diagonal(&m,&e);
      assert(res != 0);
      det_man[j] = m;
      det_exp[j] = e;
    }
  }
}

/*****
 * double clusters::EM_XMM(int numex, int clust)
 * renvoie (X-mu)^T Sigma^{-1} (X-mu) pour le cluster clust et l'exemple numex
 *****/
double clusters::EM_XMM(int numex, int clust){
  double ret=0, t;
  int i,j;
  double* temp=new (double) [dimension];
  double val;

  for (i=0;i<dimension;i++){
    val=0;
    for (j=0;j<dimension;j++){
      t = pointSet->getCoordPoints(numex,j) - getCoordCenter(clust,j); //getInvSigma(j,i,clust);
      val += t*getInvSigma(j,i,clust);
    }
    temp[i] = val;
  }
  
  for (i=0;i<dimension;i++){
    t = pointSet->getCoordPoints(numex,i) - getCoordCenter(clust,i);
    ret += t*temp[i];
  }

  /*if (ret<0.003){
    for (i=0;i<dimension;i++){
    printf("temp[%d]=%g | invSigma=%g | point=%d | center=%g\n",i,temp[i],getInvSigma(i,i,clust),pointSet->getCoordPoints(numex,i),getCoordCenter(clust,i));
    }
    printf("ret=%g\n",ret);
    }*/

  delete [] temp;
  return ret;
}

/*****
 * double clusters::EM_Pointwise_Gaussian(int numex, int clust)
 * renvoie la valeur de densite pour l'exemple numex
 * et le cluster clust
 *****/
double clusters::EM_Pointwise_Gaussian(int numex, int clust){
  double expo = EM_XMM(numex, clust);
  double den, det;
  double val;
  det = exp(log(det_man[clust])+det_exp[clust]*log(10));
  den = puissance((2*PI),dimension);
  den = sqrt(den);
  den *= sqrt(det);
  val = exp(-expo/2.0)/den;

  if (expo>BCP)
    val = TOUPITI;

  if (den==0)
    val = 1/TOUPITI;

  /*if (val<0){
    printf("expo = %g, exp(-expo/2)=%g, den=%g, det=%g || val=%g \n",expo,exp(-expo/2.0),den,det,val);
    printf("egalite %d",exp(-expo/2.0) == 0);
    varcovar[clust]->affiche();
    inverseVarcovar[clust]->affiche();
    }*/

  assert(val>=0);

  return val;
}

/*****
 * void clusters::EM_update_membership(void)
 * compute posterior probabilities for (W)EM
 *****/
void clusters::EM_update_membership(void)
{
  int j,k,l,nv;
  double num, den, ratio;
  for (j=0;j<NP;j++){
    den = 0;
    nv = 0;
    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	nv++;
	den += (EM_Pointwise_Gaussian(j,k)*mixingProp[k]);
      }
    }
    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	num = EM_Pointwise_Gaussian(j,k)*mixingProp[k];
	if (den == 0){
	  assert (num==0);
	  ratio = 0;
	}
	else{
	  ratio = num / den;
	}
	if (!(ratio >= 0)){
	  printf("%d : num=%g, den=%g, Gaussian=%g, mixing=%g\n",j,num,den,EM_Pointwise_Gaussian(j,k),mixingProp[k]);
	  for (l=0;l<KC;l++){
	    if (getClusterValide(l)==VRAI){
	      printf("** %d : Gaussian=%g, mixing=%g\n",l,EM_Pointwise_Gaussian(j,l),mixingProp[l]);
	    }
	  }
	}
	assert(ratio >= 0);
	putMembership(k,j,ratio);
      }
      else{
	putMembership(k,j,0);
      }
    }

    if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
    assert(nv>0);
  }
}

/*****
 * void clusters::EM_update_mixing_proportions(void)
 * update mixing proportions for (W)EM
 *****/
void clusters::EM_update_mixing_proportions(void)
{
  int j,k,nv=0;
  double val;
  for (k=0;k<KC;k++){
    if (getClusterValide(k)==VRAI){
      nv++;
      val = 0;
      for (j=0;j<NP;j++){
	val += getMembership(k,j);
      }
      val /= (double) NP;
      
      /*if(val<0){
	for (j=0;j<NP;j++){
	  if (getMembership(k,j)<0) printf("%d -> %g\n",j,getMembership(k,j));
	}
	}*/

      mixingProp[k] = val;
    }
  }
  assert(nv>0);
}

/*****
 * double clusters::EM_Pointwise_Loss(int p)
 * calcule et renvoie le cout EM pour un point
 * ATTENTION: tous les parametres sont supposes mis a jour
 *****/
double clusters::EM_Pointwise_Loss(int p)
{
  int k;
  double v=0,w1,w2;
  int nv=0;

  for (k=0;k<KC;k++){
    if (getClusterValide(k)==VRAI){
      w1 = 0.5*(dimension*log(2*PI) + log(det_man[k])+det_exp[k]*log(10));
      w2 = 0.5*EM_XMM(p, k);
      v+= (w1 + w2);
      nv++;
    }
  }
  v/=(double) nv;

  if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
  assert(nv>0);

  return v;
}

/*****
 * void clusters::Compute_L_KMN_EM(void)
 * calcule min -log p, qui est l'equivalent du cout KMN pour EM,
 * sauf qu'on prend en consideration les sigmas
 *****/
void clusters::Compute_L_KMN_EM(void)
{
  L_KMN=KMN_Loss_EM();
}

/*****
 * double clusters::KMN_Loss_EM(void)
 * calcule et renvoie le cout KMN des KMeans pour EM
 *****/
double clusters::KMN_Loss_EM(void)
{
  double cout =0;
  double cco;
  int i;
  for (i=0;i<NP;i++){
    cco=Pointwise_KMN_Loss_EM(i);
    cout+=cco;
  } 
  cout /= (double) NP;
  return cout;
}

/*****
 * double clusters::Pointwise_KMN_Loss_EM(int p)
 * calcule et renvoie le cout KMN EM des KMeans pour un point
 *****/
double clusters::Pointwise_KMN_Loss_EM(int p)
{
  double loss, lossmin, logexpo, logdet, logden;
  BOOL trouve;
  int k, nc, nval;

  trouve=FAUX;
  nc=-1;
  nval=0;
  for (k=0;k<KC;k++){
    if (getClusterValide(k)==VRAI){
      nval++;
      logexpo = 0.5*EM_XMM(p, k);
      logdet = log(det_man[k])+det_exp[k]*log(10);
      logden = 0.5*(logdet+dimension*log(2*PI));
      loss = logexpo + logden;
      
      if ( (trouve==FAUX) || (loss<lossmin) ) {
	nc=k;
	lossmin = loss;
	trouve=VRAI;
      }
    }
  }

  if (nval == 0) printf("\n\nATTENTION :: Aucun Cluster !!\n");
  
  assert(nc!=-1);

  return lossmin;
}

/*****
 * void clusters::Generic_Expectation_Maximization(int nstep, int typec, BOOL diagonal)
 * algorithme (weighted) EM, 2005
 * typec = 0 -> la valeur de c est fixee par la nouvelle formule
 * typec = 1 -> la valeur de c est calculee par dichotomies
 * typec = 2 -> c'est l'algo EM traditionnel qui est execute
 *  >> si diagonal == 1, on considere des Sigmas (mat de variance / covariance) diagonales
 *  >> sinon, c'est le bordel, n'ayons pas peur des mots
 *****/
void clusters::Generic_Expectation_Maximization(int nstep, int typec, BOOL diagonal){
  assert(typec != 1);

  if (typec == 0){
    strcpy(smethod,"WEM_C_Constant");
  }else if (typec == 1){
    strcpy(smethod,"WEM_C_Dicho");
  }else if (typec == 2){
    strcpy(smethod,"EM");
  }

  if (diagonal==1){
    strcat(smethod,"_Diagonal_Sigmas");
  }else{
    strcat(smethod,"_Arbitrary_Sigmas");
  }

  sauvegarde_im_theorique();

  char ns[255];
  if (typec == 0)
    strcpy(ns,"Weighted_Expectation_Maximization_C_Exact");
  else if (typec == 1)
    strcpy(ns,"Weighted_Expectation_Maximization_C_Dicho");
  else if (typec == 2)
    strcpy(ns,"Expectation_Maximization");

  int i, step;
  double dmaxt, dplust, dmoinst, lc, abl, ct, Z;
  double prop;

  double* Log_Loss_Init=new (double) [NP];
  double* Average_Log_Loss=new (double) [NP];
  double fraction;

  for (i=0;i<NP;i++) weights[i]=(double)1/(double) NP;

  EM_init_mixingProp();
  EM_init_membership();
  EM_crunch_Sigmas(diagonal);

  for (step=0;step<nstep;step++){

    EM_update_membership();

    if (step==0){
      for (i=0;i<NP;i++){
	Log_Loss_Prec[i] = EM_Pointwise_Loss(i);
	Log_Loss_Init[i] = Log_Loss_Prec[i];
      }
    }
    else{
      for (i=0;i<NP;i++){
	Log_Loss_Cour[i] = Pointwise_HM_Loss(i);	
      }
    }

    if (step==0){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Init[i];
    }else if (step<nstep-1){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Cour[i];
    }

    if ( (step>0) && ( typec < 2) )
      ct=weights_abstract(Log_Loss_Prec, Log_Loss_Cour, typec, step);

    //calculs divers
    Compute_L_KMN_EM();

    if ( (step==0) || (L_KMN < Min_L_KMN) )
      Min_L_KMN = L_KMN;

    if (step==nstep-1){ //ATTENTION: CHANGEMENT !!
      fraction=0;
      for (i=0;i<NP;i++){
	if (Log_Loss_Init[i] <= Log_Loss_Cour[i])
	  fraction++;
      }
      fraction/=(double) NP;
      //if (fraction < Min_Prop)
      Min_Prop = fraction;
    }

    //sauvegardes diverses
    
    if (step>=0){
      if (typec < 2)
	if (ct<0) sauvegarde_param('-',ns,step);
	else if (ct==0) sauvegarde_param('z',ns,step);
	else sauvegarde_param('+',ns,step);
      else
	sauvegarde_param('.',ns,step);
    }

    if (step>0)
      for (i=0;i<NP;i++) Log_Loss_Prec[i] = Log_Loss_Cour[i];

    if (step==nstep-1)
      Prop_Missed_Fin_Algo = count_prop_missed_clusters();

    //recalcul des centres et sigmas, distibs
    calcule_centres_clusters();
    EM_crunch_Sigmas(diagonal);
    EM_update_mixing_proportions();

    if (step%STEP_FOR_SAVING_IMG==0)
      sauvegarde_im(step);
    printf(".");
    fflush(stdout);
  }

  for (i=0;i<NP;i++)
    if ( (i==0) || (Log_Loss_Cour[i]>Max_Loss) )
      Max_Loss = Log_Loss_Cour[i];

  for (i=0;i<NP;i++){
    Average_Log_Loss[i]/=(double) (nstep-1);
  }

  fraction=0;
  for (i=0;i<NP;i++){
    if (Log_Loss_Cour[i] >= Average_Log_Loss[i])
      fraction++;
  }
  fraction/=(double) NP;
  Log_Prop = fraction;

  printf("]\n");
  sauvegarde_scatter();

  //for (i=0;i<KC;i++) printf("[%d] ",clusterValide[i]);

  delete [] Log_Loss_Init;
  delete [] Average_Log_Loss;
}

/***********************************************************************************
 *
 * FUZZY C MEANS
 *
 *****/

/*****
 * void clusters::calcule_centres_clusters_FCM(int a)
 * recalcule les centres de clusters, sur la base du vecteur de poids
 * pour les exemples, pour les Fuzzy C Means
 *****/
void clusters::calcule_centres_clusters_FCM(int a)
{
  int i,j,k;
  double sav, sap, pm;
  double* denomK=new (double) [KC];
  double* calcvalid=new (double) [KC];

  for (j=0;j<KC;j++){
    denomK[j]=0;
    calcvalid[j]=0;
  }
  for (i=0;i<KC;i++)
    for (j=0;j<dimension;j++) 
      putCoord(centersCoordFromClusters,i,j,0);
    
  for (i=0;i<NP;i++){
    for (k=0;k<KC;k++){
      //printf("Calcule_Centre_Clusters: (%d,%d) : %d,(W=%g,M=%g)\n",i,k,getClusterValide(k),weights[i],getMembership(k,i));
      if (getClusterValide(k)==VRAI){
	pm = pow (getMembership(k,i), (double) a);
	calcvalid[k]+=pm;
	denomK[k]+=weights[i]*pm;
	for (j=0;j<dimension;j++){
	  sav = getCoord(centersCoordFromClusters,k,j);
	  sap = sav + ( (double) pointSet->getCoordPoints(i,j)*(weights[i]*pm));
	  putCoord(centersCoordFromClusters,k,j,sap);
	}
      }
    }
  }
  
  for (i=0;i<KC;i++){
    if (getClusterValide(i)==VRAI){
      if (calcvalid[i]>0){
	for (j=0;j<dimension;j++){
	  sap = (getCoord(centersCoordFromClusters,i,j)/denomK[i]);
	  putCoord(centersCoordFromClusters,i,j,sap);
	}
      }
      else{
	putClusterValide(i,FAUX);
	printf("[Removing cluster %d]",i);
      }
    }
  }

  delete [] calcvalid;
  delete [] denomK;
}


/*****
 * void clusters::membership_FCM(int a)
 * affecte les points aux clusters, sachant que les centres sont choisis
 * ATTENTION: la methode est celle du Fuzzy C Means clustering
 * a = parametre m dans l'algo
 *****/
void clusters::membership_FCM(int a)
{
  int j,k,l,nv;
  double contribcour, sumcont, num, denom, A, B, puiss;
  double suminv;

  for (j=0;j<NP;j++){
    sumcont = 0;
    nv = 0;

    //on calcule la somme des inverses
    suminv = 0;
    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	num = sqrt(distance_pt_centre(j,k,centersCoordFromClusters));
	if (num < EPSILON_FCM)
	  num = EPSILON_FCM;
	A = num;
	B = (double) 2 / (double) (a - 1);
	num = pow(A, B);
	num = 1 / num;
	suminv += num;
      }
    }

    for (k=0;k<KC;k++){
      if (getClusterValide(k)==VRAI){
	nv++;

	num = sqrt(distance_pt_centre(j,k,centersCoordFromClusters));
	A = num;
	B = (double) 2 / (double) (a - 1);
	num = pow(A, B);
	contribcour = num * suminv;
	putMembership(k,j,contribcour);
      } else {
	putMembership(k,j,0);
      }
    }

    /*for (k=0;k<KC;k++){
      //printf("[%d %d] %d \n", j, k, getClusterValide(k));

      if (getClusterValide(k)==VRAI){
	nv ++;

	num = sqrt(distance_pt_centre(j,k,centersCoordFromClusters));
	for (l=0;l<KC;l++){
	  if (getClusterValide(l)==VRAI){
	    denom = sqrt(distance_pt_centre(j,l,centersCoordFromClusters));
	    if (denom < EPSILON_FCM)
	      denom = EPSILON_FCM;
	    A = num / denom;
	    B = (double) 2 / (double) (a - 1);
	    puiss = pow(A, B);
	    contribcour += puiss;
	  }
	}
	if (contribcour < EPSILON_FCM)
	  contribcour = EPSILON_FCM;
	contribcour = 1 / contribcour;
	putMembership(k,j,contribcour);
      } else {
	putMembership(k,j,0);
      }
      }*/

    if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
    assert(nv>0);
  }
}

/*****
 * double clusters::Pointwise_FCM_Loss(int p, int a)
 * calcule et renvoie le cout Fuzzy C Means pour un point p
 *****/
double clusters::Pointwise_FCM_Loss(int p, int a)
{
  int k, nv=0;
  double v=0;
  double m, da;

  for (k=0;k<KC;k++){
    if (getClusterValide(k)==VRAI){
      m = getMembership(k,p);
      da = (double) a;
      m = pow(m, da);
      v+= m*distance_pt_centre(p,k,centersCoordFromClusters);
      nv ++;
    }
  }

  if (nv==0) printf("Pas assez de points ou trop de clusters : arret\n");
  assert(nv>0);

  return v;
}

/*****
 * double clusters::Fuzzy_C_Means_Loss(int a)
 * calcule et renvoie le cout Fuzzy C Means moyen
 *****/
double clusters::Fuzzy_C_Means_Loss(int a)
{
  double cout =0;
  double cco;
  int i;
  for (i=0;i<NP;i++){
    cco=Pointwise_FCM_Loss(i,a);
    cout+=cco;
  } 
  //cout = sqrt(cout);
  cout /= (double) NP;
  return cout;
}


/*****
 * void clusters::Generic_Fuzzy_C_Means(int nstep, int typec)
 * algorithme des (weighted) Fuzzy C-means, 2005
 * typec = 0 -> la valeur de c est fixee par la nouvelle formule
 * typec = 1 -> la valeur de c est calculee par dichotomies
 * typec = 2 -> c'est l'algo des Harmonic-means traditionnel qui est execute
 *****/
void clusters::Generic_Fuzzy_C_Means(int nstep, int typec){

  int a = M_FCM; //coefficient m dans les FCM

  if (typec == 0){
    strcpy(smethod,"FCM_C_Constant");
  }else if (typec == 1){
    strcpy(smethod,"FCM_C_Dicho");
  }else if (typec == 2){
    strcpy(smethod,"FCM");
  }

  //sauvegarde_im_theorique();

  char ns[255];
  if (typec == 0)
    strcpy(ns,"Fuzzy_C_Means_C_Exact");
  else if (typec == 1)
    strcpy(ns,"Fuzzy_C_Means_C_Dicho");
  else if (typec == 2)
    strcpy(ns,"Fuzzy_C_Means");

  int i, step;
  double dmaxt, dplust, dmoinst, lc, abl, ct, Z;
  double prop;

  double* Log_Loss_Init=new (double) [NP];
  double* Average_Log_Loss=new (double) [NP];
  double fraction;

  for (i=0;i<NP;i++) weights[i]=(double)1/(double) NP;

  for (step=0;step<nstep;step++){

    membership_FCM(a);

    if (step==0){
      for (i=0;i<NP;i++){
	Log_Loss_Prec[i] = Pointwise_FCM_Loss(i,a);
	Log_Loss_Init[i] = Log_Loss_Prec[i];
      }
      //divide_loss(Log_Loss_Prec);
      //divide_loss(Log_Loss_Init);
    }
    else{
      for (i=0;i<NP;i++){
	Log_Loss_Cour[i] = Pointwise_FCM_Loss(i,a);	
      }
      //divide_loss(Log_Loss_Cour);
      for (i=0;i<NP;i++){
	//printf("[%d] Prec = %g, Cour = %g /// %d \n",i,Log_Loss_Prec[i],Log_Loss_Cour[i],(Log_Loss_Prec[i]>Log_Loss_Cour[i]) ? 1 : 0);
      }
    }

    if (step==0){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Init[i];
    }else if (step<nstep-1){
      for (i=0;i<NP;i++)
	Average_Log_Loss[i]+= Log_Loss_Cour[i];
    }

    if ( (step>0) && ( typec < 2) )
      ct=weights_abstract(Log_Loss_Prec, Log_Loss_Cour, typec, step);

    //calculs divers
    Compute_L_KMN_Without_GetMembership();//Compute_L_HM();
    if ( (step==0) || (L_KMN < Min_L_KMN) )
      Min_L_KMN = L_KMN;

    if (step==nstep-1){ //ATTENTION: CHANGEMENT !!
      fraction=0;
      for (i=0;i<NP;i++){
	if (Log_Loss_Init[i] <= Log_Loss_Cour[i])
	  fraction++;
      }
      fraction/=(double) NP;
      //if (fraction < Min_Prop)
      Min_Prop = fraction;
    }

    //sauvegardes diverses
    
    if (step>=0){
      if (typec < 2)
	if (ct<0) sauvegarde_param('-',ns,step);
	else if (ct==0) sauvegarde_param('z',ns,step);
	else sauvegarde_param('+',ns,step);
      else
	sauvegarde_param('.',ns,step);
    }

    if (step>0)
      for (i=0;i<NP;i++) Log_Loss_Prec[i] = Log_Loss_Cour[i];

    if (step==nstep-1)
      Prop_Missed_Fin_Algo = count_prop_missed_clusters();

    //recalcul des centres
    calcule_centres_clusters_FCM(a);

    if (step%STEP_FOR_SAVING_IMG==0)
      sauvegarde_im(step);
    printf(".");
    fflush(stdout);
  }

  for (i=0;i<NP;i++)
    if ( (i==0) || (Log_Loss_Cour[i]>Max_Loss) )
      Max_Loss = Log_Loss_Cour[i];

  for (i=0;i<NP;i++){
    Average_Log_Loss[i]/=(double) (nstep-1);
  }

  fraction=0;
  for (i=0;i<NP;i++){
    if (Log_Loss_Cour[i] >= Average_Log_Loss[i])
      fraction++;
  }
  fraction/=(double) NP;
  Log_Prop = fraction;

  printf("]\n");
  sauvegarde_scatter();

  delete [] Log_Loss_Init;
  delete [] Average_Log_Loss;
}
