/*****
* cluster.h : header
*****/

/*****
* fichiers inclus
*****/

#include"types_communs.h"
#include"additional.h"
#include<stdlib.h>
#include<stdio.h>
#include<time.h>
#include<assert.h>
#include<math.h>
#include<string.h>

/*****
* definition de macros
*****/
#define min(x,y) ((x)>(y) ? (y) : (x))
#define max(x,y) ((x)>(y) ? (x) : (y))
#define abs(x)  ((x)<0 ? (-x) : (x))

//debuggage
#define _DEBUG

#ifdef _DEBUG
#define DEBUG(x)    printf(x)
#define DEBUGII(x,y) printf(x,y)
#define ASSERT(x)   assert(x)
#else
#define DEBUG(x)      /*  */
#define DEBUGII(x,y)  /*  */
#define ASSERT(x)     /*  */
#endif

/*****
* definition de constantes
*****/

#ifndef NULL
#define NULL   (0)
#endif

#define PI_MATH 3.141592

/*****
* types predefinis
*****/

typedef enum BOOL {FAUX, VRAI};

/*****
 * classe couleur
 * faite pour declarer eventuellement des constantes de couleurs
 * je m'eclate ! [sic]
 *****/
class couleur
{
 private:
  int r;
  int g;
  int b;
  int a;

 public:
  couleur(int R, int G, int B, int A){
    r=R;
    g=G;
    b=B;
    a=A;
  }

  int getR(void) {return r;}
  int getG(void) {return g;}
  int getB(void) {return b;}
  int getA(void) {return a;}
};

/*****
 * classe imageC
 * classe gerant les images (creation, projection de clustering, sauvegarde, etc. )
 *****/
class imageC
{
 private:
  int Wi; //largeur
  int Le; //hauteur
  pixel** bm; //tableau de pixels de l'image
  const char* filename; //nom du fichier associe

public:
  imageC(const char *file, int width, int length); // constructeur
  ~imageC(void);// destructeur

  int vWi(void) {return Wi;}
  int vLe(void) {return Le;}

  void point(int x, int y, int r, int g, int b, int a, int T){
    int i,j;
    for (i=y-(T/2);i<=y+(T/2);i++) 
      for (j=x-(T/2);j<=x+(T/2);j++){
	if ( (i>=0) && (i<Le) && (j>=0) && (j<Wi) ){
	  bm[i*Wi+j]->r=r;
	  bm[i*Wi+j]->g=g;
	  bm[i*Wi+j]->b=b;
	  bm[i*Wi+j]->a=a;
	}
      }
  }

  void cercle(int x, int y, int R, int r, int g, int b, int a, int T)
    {
      int i;
      double alpha;
      for(i=0;i<100;i++){
	alpha=2.0*i*PI_MATH/100.0;
	point(x+(int) (R*cos(alpha)), y+(int) (R*sin(alpha)), r, g, b, a, T);
      }
    }

 void ligne(int x0, int y0, int x1, int y1, int r, int g, int b, int a, int T)
   {
     int dy = y1 - y0;
     int dx = x1 - x0;
     int stepx, stepy;
 
     if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
     if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
     dy <<= 1;                                                  
     dx <<= 1;                                                  
     
     point(x0, y0, r, g, b, a, T);
     if (dx > dy) {
       int fraction = dy - (dx >> 1);                         
       while (x0 != x1) {
	 if (fraction >= 0) {
	   y0 += stepy;
	   fraction -= dx;                                
	 }
	 x0 += stepx;
	 fraction += dy;                                    
	 point(x0, y0, r, g, b, a, T);
       }
     } else {
       int fraction = dx - (dy >> 1);
       while (y0 != y1) {
	 if (fraction >= 0) {
	   x0 += stepx;
	   fraction -= dy;
	 }
	 y0 += stepy;
	 fraction += dx;
	 point(x0, y0, r, g, b, a, T);
       }
     } 
   }


  void sauvegarde(void); //sauvegarde l'image
  int saveRGBtiff(const char *filename, const char *soft, const char *desc,
		  unsigned int *data, short photometric, int width, int height); // annexe sauvegarde
};

/*****
 * classe tabPoints
 * classe gerant les pointsets
 *****/

class tabPoints
{
private:

  //variables liees aux points

  int dimension;
    // dimension du probleme
  int NP;
    // nbre total de points;

  int* coordsPoints;
    // tableau ( une dim, attention  la gestion des coords) des coords
    // [... Coord 1...][... Coord2...]...[... Coord-dimension...] = [X][Y][Z]...

  int*maxC, *minC;
    // coordonnees min/max par dimension

  int *Nearest_Neighbor;
    // Tableau donnant les K plus proches voisins d'un point donne
    // tableau a 1 dim: [ ... NN1 ... ][ ... NN2 ... ]...
  int NN; //nb de NN

  //variables liees aux clusters theoriques

  int generation;
    // methode de generation des clusters thoriques

  int balance;
    // methode de choix des proportions des clusters dans les donnees

  int KC_Theorique;
    // nombre de clusters theoriques

  int* averageCoordCentres;
    // tableau ( une dim, attention  la gestion des coords) des coords estimes des centres de clusters

  double* sigmaClusters;
  // tableau des ecarts-type

  double* echantillonnage;
    // tab de taille KC_Theorique donnant la proba de tirer un point
    // de chaque cluster

  int* clusterTheorique;
    // numero de cluster a partir duquel a ete genere l'exemple i

public:
  tabPoints(int dim, int gen, int bal, int n, int k); // constructeur
  ~tabPoints(void);// destructeur

  int vNP(void) {return NP;}
  int vdim(void) {return dimension;}
  int vKC_th(void) {return KC_Theorique;}
  int vMax(int c) {return maxC[c];}
  int vMin(int c) {return minC[c];}
  int vgeneration(void) {return generation;}
  int vbalance(void) {return balance;}
  int vcluster(int n) {return clusterTheorique[n];}

  void putCoordPoints(int p, int nc, int val) {coordsPoints[NP*nc+p]=val;}
  int getCoordPoints(int p, int nc){return coordsPoints[NP*nc+p];}

  void putCoordCentre(int c, int nc, int val){averageCoordCentres[KC_Theorique*nc+c]=val;}
  int getCoordCentre(int c, int nc){return averageCoordCentres[KC_Theorique*nc+c];}
  
  void putNeighbor(int p, int num, int pn) {Nearest_Neighbor[p*NN+num]=pn;}
  int getNeighbor(int p, int num){return Nearest_Neighbor[p*NN+num];}

  void affiche(void);
  // affiche les coords des points et leur cluster theorique

  void claque_im(imageC* im, BOOL colors);
  // place les points dans l'image im selon les deux premiere coords
  // si colors==TRUE, on met les couleurs theoriques

  //Methodes de generation aleatoire des clusters
  void generate(BOOL Harmonic);
  // generation aleatoire des points
  // Harmonic = BOOLEEN : si VRAI, on divise tout par NORMALISATION_CONST (moyennes et ecart-types)

  double GaussianSigma(double ave, double sigma);
  // tirage selon N(ave,sigma)

  void generateAveSigma(BOOL Harmonic);
  // generation aleatoire des esperances et sigma
  // Harmonic = BOOLEEN : si VRAI, on divise tout par NORMALISATION_CONST (moyennes et ecart-types)

  void stretchClusters(void);
  // reduit les coordonnees pour que les points tiennent dans un hypercube
  // de cote COORD_MAX

  void generateGaussianClusters(void);
  // genere des clusters gaussiens en ND

  void generateCubicClusters(void);
  // genere des clusters Cubiques en ND, dont chaque composante
  // des points est uniformement choisie entre M-2S et M+2S

  void generateRingGaussianClusters(void);
  // genere des clusters gaussiens sur des anneaux en ND

  void generateGridGaussianClusters(void);
  // genere des clusters gaussiens en 2D sur une grille
  // (jeu de donnees de type BIRCH)

  //Methodes de calcul divers

  int puiss(int a, int b);
  // renvoie un int a^b

  int plus_pt_int(int a, int b);
  // renvoie le plus petit entier c tel que c^a >= b

  double distance(int p1, int p2){
    double v=0;
    int j;
    for (j=0;j<dimension;j++) v+=(double) ((getCoordPoints(p1,j)-getCoordPoints(p2,j))
				*(getCoordPoints(p1,j)-getCoordPoints(p2,j)));
    v=sqrt(v);
    return v;
  }

  void initNN(void);
  // initialise les NN
};


/*****
 * classe clusters
 * classe gerant les algos de clustering
 *****/

class clusters
{
 private:

  tabPoints* pointSet;
    // pointSet associe

  int NP;
    // recopie de NP (evite manips)
  int dimension;
    // idem
  int KC;
    // nbre de clusters a trouver

  double* membershipClusters;
    // appartenance de chaque exemple a chaque cluster
    // attention: tab a 1 dimension: [ appartenance premier exemple ] [ appartenance 2eme ex ] [ ... ]

  double* weights;
    // tableau des poids des points;
  double maxWeight;
    // poids max sur les exemples
  double minWeight;
    // poids min sur les exemples

  double* Log_Loss_Prec;
    // couts precedents (au depart, couts initiaux)
  double* Log_Loss_Cour;
    // couts courants (calcules pour optimise c, puis mis-a-jour avec les nouv centres)

  // Variables associees aux nouveaux clusters
  double* centersCoordFromClusters;
    // centres calcules a partir des points d'un cluster
    // attention: tab a 1 dimension: [ premiere coord sur tous les clusters ] [ 2eme coord ] [ ... ]

  //EM//
  //Matrices de variances / covariances (seulement EM)
  matrice** varcovar;
  matrice** inverseVarcovar;
  double* det_man;
  int* det_exp;
  double* mixingProp;


  BOOL* clusterValide;
    // contient false ssi le cluster ne contient plus d'element.

  double L_KMN;
    // Cout de KMN

  double Min_L_KMN;
    // Cout minimal KMN durant les iterations

  double Min_Prop;
    // Proportion minimale de points dont la fonction de cout ne decroit pas

  double Log_Prop;
    // Proportion minimale de points dont le log de la fonction de cout
    // n'est pas inferieur a la moyenne des log sur les iterations

  double Prop_Missed_Fin_Algo;
    // Proportion de clusters theoriques non touches en fin d'algo
    // (a la difference des autres Min_* valeurs, on ne peut arreter experimentalement
    // l'algo a la prop minimale, donc on prend celle en fin d'algo.

  double Max_Loss;
    // cout maximal sur un point

  //////////////////////////////////////////////////////////////////////////
  // divers

  char* sfile;
    // nom generique du fichier de sauvegarde

  char* smethod;
    // nom de la methode adoptee

  int t_im;
    // width=length de l'image qui comportera la config

 public:
  clusters(tabPoints* tab, int K, char* nf, int t_i, int S, int W, int C, int I, BOOL INIT); // constructeur
  ~clusters(void);// destructeur

  void reinitialise_clustering(clusters* cap);
    // initialise le clustering a partir d'un autre clustering

  void copie_cluster_centers(clusters* cap);
    // copie les centres de clusters choisis pour cap dans le receveur

  void putSigma(int row, int column, int clust, double v){varcovar[clust]->putCoord(row,column,v);}
  double getSigma(int row, int column, int clust){return varcovar[clust]->getCoord(row,column);}

  void putInvSigma(int row, int column, int clust, double v){inverseVarcovar[clust]->putCoord(row,column,v);}
  double getInvSigma(int row, int column, int clust){return inverseVarcovar[clust]->getCoord(row,column);}

  void putCoord(double* t, int numero_cluster, int numero_coord, long double val){t[KC*numero_coord+numero_cluster]=val;}
  double getCoord(double* t, int numero_cluster, int numero_coord){return t[KC*numero_coord+numero_cluster];}

  double getCoordCenter(int numero_cluster, int numero_coord){return getCoord(centersCoordFromClusters, numero_cluster,  numero_coord);}

  double getMembership(int numero_cluster, int numero_ex){return membershipClusters[KC*numero_ex + numero_cluster];}
  void putMembership(int numero_cluster, int numero_ex, double val){membershipClusters[KC*numero_ex + numero_cluster]=val;}

  BOOL getClusterValide(int numero_cluster){return clusterValide[numero_cluster];}
  void putClusterValide(int numero_cluster, BOOL val){clusterValide[numero_cluster]=val;}

  double vL_KMN(void) {return L_KMN;}
  double vMax_Loss(void) {return Max_Loss;}
  double vMin_L_KMN(void) {return Min_L_KMN;}
  double vMin_Prop(void) {return Min_Prop;}
  double vLog_Prop(void) {return Log_Prop;}
  double vProp_Missed_Fin_Algo(void) {return Prop_Missed_Fin_Algo;}

  ///////////////////////////////////////////////////////////////////////////
  // methodes rattachees au dessin des clusters

  void claque_im(imageC* im);
    // place les clusters empiriques dans l'image im selon les deux premiere coords

  void put_convex_hulls(imageC* im);
    // place les enveloppes convexes des clusters empiriques dans l'image im selon les deux premiere coords

  double isLeft(double p0x, double p0y, double p1x, double p1y, double p2x, double p2y);
    // return >0 for P2 left of the line through P0 and P1
    // =0 for P2 on the line
    //           <0 for P2 right of the line

  void sort(double* Px, double* Py, int n);
    // ordonne lexico P selon un tri hyper lent [beurk]

  int chainHull_2D(double* Px, double* Py, int n, double* Hx, double* Hy);
    //     Input:  P[] = an array of 2D points
    //                   presorted by increasing x- and y-coordinates
    //             n = the number of points in P[]
    //     Output: H[] = an array of the convex hull vertices (max is n)
    //     Return: the number of points in H[]

  ////////////////////////////////////////////////////////////////////////////
  // Methodes rattachees aux sauvegardes

  char * itoa(int i);
    // passe l'entier i dans une chaine de char

  void sauvegarde_scatter(void);
    // sauvegarde L_KMN

  void sauvegarde_param(char a, char* entete, int iter);
    // sauvegarde L_KMN et Variabilite
    // entete est la chaine placee en debut de fichier

  void sauvegarde_costs_against(clusters* wcap);
    // sauvegarde L_KMN pour le cluster recevant et le cluster en parametre,
    // et calcule le gain relatif de L_KMN pour wcap par rapport au clusters
    // recevant.

  void sauvegarde_props_against(clusters* wcap);
    // sauvegarde Min_Prop pour le cluster recevant et le cluster en parametre,
    // et calcule le gain relatif de Min_Prop pour wcap par rapport au clusters
    // recevant.

  void sauvegarde_average_props_against(clusters* wcap);
    // sauvegarde Log_Prop pour le cluster recevant et le cluster en parametre,
    // et calcule le gain relatif de Log_Prop pour wcap par rapport au clusters
    // recevant.
    // Log_Prop est la proportion de points pour lesquels le log de la fonction
    // de cout a la fin est inferieur a la moyenne de tous les precedents: c'est
    // donc aussi la proportion de points pour lesquels le C (dans EXP(-C)) est
    // superieur a la moyenne des precedents: selon la theorie, cette quantite pourrait
    // decroitre exponentiellement

  void sauvegarde_missed_against(clusters* wcap);
    // sauvegarde Prop_Missed_Fin_Algo pour le cluster recevant et le cluster en parametre,
    // et calcule le gain relatif de Prop_Missed_Fin_Algo pour wcap par rapport au clusters
    // recevant.

  void sauvegarde_max_loss_against(clusters* wcap);
    // sauvegarde Max_Loss pour le cluster recevant et le cluster en parametre,
    // (calcule en fin d'iteration)
    // et calcule le gain relatif de Max_Loss pour wcap par rapport au clusters
    // recevant.

  void sauvegarde_im(int iter);
    // sauvegarde l'image courante

  void sauvegarde_im_theorique(void);
    // sauvegarde le clustering theorique

  void sauvegarde_hits(char* entete, int it);
    // sauvegarde le nombre de clusters theoriques touches

  int count_target(void);
    // renvoie le combre de clusters theoriques qui comptent au moins 1 centre
    // estime

  double count_prop_missed_clusters(void);
    // renvoie la proportion de clusters theoriques qui ne comptent aucun
    // centre experimental dans leur cellule de Voronoi

  ///////////////////////////////////////////////////////////////////////////
  // methodes generiques

  void put_dist(int v);
    // remplace la distance

  double distance_pt_centre(int pt, int centre, double *tabCentres);
    // renvoie la distance euclidienne entre un pt de pointSet et un centre

  double KMN_Loss_KMeans(void);
    // calcule et renvoie le cout KMN des KMeans

  double KMN_Loss_KMeans_Without_GetMembership(void);
    // calcule et renvoie le cout KMN des KMeans
    // Mais n'utilise pas la fonction getMembership(): on recalcule completement
    // l'appartenance du point

  double Pointwise_KMN_Loss_KMeans(int p);
    // calcule et renvoie le cout KMN des KMeans pour un point

  double Pointwise_KMN_Loss_KMeans_Without_GetMembership(int p);
    // calcule et renvoie le cout KMN des KMeans pour un point
    // Mais n'utilise pas la fonction getMembership(): on recalcule completement
    // l'appartenance du point

  double Harmonic_Means_Loss(void);
    // calcule et renvoie le cout Harmonique moyen

  double Pointwise_HM_Loss(int p);
    // calcule et renvoie le cout Harmonic Means pour un point

  void Compute_L_KMN(void);
    // calcule L_KMN

  void Compute_L_KMN_Without_GetMembership(void);
    // calcule L_KMN
    // Mais n'utilise pas la fonction getMembership(): on recalcule completement
    // l'appartenance du point

  void Compute_L_HM(void);
    // calcule L_HM

  ///////////////////////////////////////////////////////////////////////////
  // methodes de bases rattachees au clustering

  void init_clusters_centers_Forgy(void);
    // init aleatoire des centres selon E. Forgy

  void init_clusters_centers_Random(void);
    // init aleatoire des centres selon Random Partition

  void affecte_points_clusters_Hard_Membership(void);
    // affecte les points aux clusters, sachant que les centres sont choisis
    // ATTENTION: la methode est celle du Hard Membership

  void membership_KHM(void);
    // affecte les points aux clusters, sachant que les centres sont choisis
    // ATTENTION: la methode est celle du Harmonic Means Clustering

  void membership_WKHM(void);
    // affecte les points aux clusters, sachant que les centres sont choisis
    // ATTENTION: la methode est celle du Weighted Harmonic Means Clustering

  double weights_abstract(double* Prec, double* Cour, int typec, int step);
    // calcule les poids des exemples dans l'abstract weighting scheme

  void weights_KHM(void);
    // calcule les poids des points
    // ATTENTION: la methode est celle du Harmonic Means Clustering

  void calcule_centres_clusters(void);
    // recalcule les centres de clusters, sur la base du vecteur de poids
    // pour les exemples, selon la methode des moindre carres

  ////////////////////////////////////////////////////////////////////////////
  // algorithmes de clustering

  void Generic_K_Means(int nstep, int typec);
    // algorithme des weighted k-means, 2005
    // typec = 0 -> la valeur de c est fixee par la nouvelle formule
    // typec = 1 -> la valeur de c est calculee par dichotomies
    // typec = 2 -> c'est l'algo des k-means traditionnel qui est execute

  void divide_loss(double* tab);
    // divise le cout par une constante, NORMALISATION_CONST

  void Generic_Harmonic_Means(int nstep, int typec);
    // algorithme des (weighted) HARMONIC k-means, 2005
    // typec = 0 -> la valeur de c est fixee par la nouvelle formule
    // typec = 1 -> la valeur de c est calculee par dichotomies
    // typec = 2 -> c'est l'algo des Harmonic-means traditionnel qui est execute

  double Z_prime(double* Prec, double* Suiv, double c);
    // derivee de Z, calculee avec c

  double c_dicho(double* Prec, double* Suiv);
    // approximation de c par dichotomies

  double c_fixe(double* Prec, double* Suiv);
    // renvoie la valeur de c sans recherche dicho, avec le nouveau resultat

  void EM_init_mixingProp(void);
    // initialise mixingProp uniforme

  void EM_init_membership(void);
    // initialise membership uniforme


  void EM_checkup(BOOL diagonal);
    // si un cluster est sous-represente, on le supprime

  void EM_approximate_checkup(BOOL diagonal);
    // si un cluster est sous-represente, on le supprime

  void EM_redispatche_membership(int clust);
    // on redispatche les memberships, sachant que clust
    // est supprime

  void EM_crunch_Sigmas(BOOL diagonal);
    // calcule Sigmas, recopie dans les inverses, et calcule les inverses

  void EM_compute_Sigmas(BOOL diagonal);
    // calcule les Sigmas
    // ATTENTION: les centres sont supposes calcules

  void EM_Sigmas_To_Inverse(void);
    // recopie les Sigmas dans les Inverses (avant de calculer les inverses)

  void EM_compute_Inverse_Sigmas(BOOL diagonal);
    // calcule les inverse des Sigmas, ainsi que det_exp et det_man
    // ATTENTION: les Sigmas sont supposes calcules !



  double EM_XMM(int numex, int clust);
    // renvoie (X-mu)^T Sigma^{-1} (X-mu) pour le cluster clust et l'exemple numex

  double EM_Pointwise_Gaussian(int numex, int clust);
    // renvoie la valeur de densite pour l'exemple numex
    // et le cluster clust

  void EM_update_membership(void);
    // compute posterior probabilities for (W)EM

  void EM_update_mixing_proportions(void);
    // update mixing proportions for (W)EM

  double EM_Pointwise_Loss(int p);
    // calcule et renvoie le cout EM pour un point
    // ATTENTION: tous les parametres sont supposes mis a jour

  void Compute_L_KMN_EM(void);
    // calcule min -log p, qui est l'equivalent du cout KMN pour EM,
    // sauf qu'on prend en consideration les sigmas

  double KMN_Loss_EM(void);
    // calcule et renvoie le cout KMN des KMeans pour EM

  double Pointwise_KMN_Loss_EM(int p);
    // calcule et renvoie le cout KMN EM des KMeans pour un point

  void Generic_Expectation_Maximization(int nstep, int typec, BOOL diagonal);
    // algorithme (weighted) EM, 2005
    // typec = 0 -> la valeur de c est fixee par la nouvelle formule
    // typec = 1 -> la valeur de c est calculee par dichotomies
    // typec = 2 -> c'est l'algo EM traditionnel qui est execute
    //   >> si diagonal == 1, on considere des Sigmas (mat de variance / covariance) diagonales
    //   >> sinon, c'est le bordel, n'ayons pas peur des mots

  /***********************************************************************************
   *
   * FUZZY C MEANS
   *
   *****/

  void calcule_centres_clusters_FCM(int a);
    // recalcule les centres de clusters, sur la base du vecteur de poids
    // pour les exemples, pour les Fuzzy C Means

  void membership_FCM(int a);
    // affecte les points aux clusters, sachant que les centres sont choisis
    // ATTENTION: la methode est celle du Fuzzy C Means clustering
    // a = parametre m dans l'algo

  double Pointwise_FCM_Loss(int p, int a);
    // calcule et renvoie le cout Fuzzy C Means pour un point p

  double Fuzzy_C_Means_Loss(int a);
    // calcule et renvoie le cout Fuzzy C Means moyen

  void Generic_Fuzzy_C_Means(int nstep, int typec);
    // algorithme des (weighted) Fuzzy C-means, 2005
    // typec = 0 -> la valeur de c est fixee par la nouvelle formule
    // typec = 1 -> la valeur de c est calculee par dichotomies
    // typec = 2 -> c'est l'algo des Harmonic-means traditionnel qui est execute

};
